import os
import logging
import sys
import collections

if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from collections.abc import Mapping
else:
    from collections import Mapping
import numpy as np
import pandas as pd
import scipy.stats as stats
import numpy.linalg as la
import time
from datetime import datetime
import pysnptools.util as pstutil
from bed_reader import get_num_threads
from pysnptools.snpreader import Bed, Pheno, SnpData, SnpReader, SnpNpz
from pysnptools.kernelreader import KernelData, KernelNpz
from pysnptools.util.mapreduce1 import map_reduce
from pysnptools.util.mapreduce1.mapreduce import _identity, _MapReduce
from pysnptools.util.mapreduce1.runner import Local
from pysnptools.util.intrangeset import IntRangeSet
from pysnptools.util import log_in_place
from pysnptools.util import format_delta
from pysnptools.util import _file_transfer_reporter
from pysnptools.snpreader import SnpMemMap
from pysnptools.standardizer import Unit
from pysnptools.snpreader import SnpGen
from pysnptools.util.filecache import LocalCache, FileCache
from fastlmm.inference.fastlmm_predictor import _snps_fixup, _pheno_fixup
from fastlmm.util.mingrid import minimize1D
from fastlmm.util.matrix.mmultfile import mmultfile_b_less_aatb, mmultfile_ata
from unittest.mock import patch


def single_snp_scale(
    test_snps,
    pheno,
    G0=None,
    covar=None,
    cache=None,
    memory_factor=1,
    output_file_name=None,
    K0=None,
    runner=None,
    min_work_count=1,
    gtg_runner=None,
    gtg_min_work_count=None,
    svd_runner=None,
    postsvd_runner=None,
    postsvd_min_work_count=None,
    test_snps_runner=None,
    test_snps_min_work_count=None,
    count_A1=False,
    clear_local_lambda=None,
    force_python_only=False,
):
    """
    Function performing single SNP GWAS using REML and cross validation over the chromosomes. Will reorder and intersect IIDs as needed.
    It gives the same results as :func:`.single_snp` but scales a little better on a single machine and has the ability to run on a cluster. (Cluster
    runs require appropriate modules for parameters ``cache`` and ``runner``.)

    To work with a G0 kernel larger than about 16K, requires a NumPy library that can do eigenvalue decompositions larger than 16K,
    for example, one that uses MKL IPL64. Alternatively, use an older version of FaST-LMM (version 0.5.\*) which includes its own version of MKL IPL64.

    Compared to :func:`.single_snp`, :func:`.single_snp_scale` always:

    * does cross validation of chromosomes (:func:`.single_snp`'s ``leave_out_one_chrom=True``)
    * use a low-rank kernel only (:func:`.single_snp`'s ``force_low_rank=True``)
    * uses exactly one kernel constructed from SNPs
    * searches for the best ``h2``. (:func:`.single_snp`'s ``h2=None``)

    :param test_snps: SNPs to test. Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__.
          If you give a string, it should be the name of a `Bed <http://fastlmm.github.io/PySnpTools/#snpreader-bed>`__-formatted file.
          For cluster runs, this and ``G0`` should be readable from any node on the cluster, for example, by using
          `DistributedBed <http://fastlmm.github.io/PySnpTools/#snpreader-distributedbed>`__ format.
    :type test_snps: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__ or a string

    :param pheno: One or more phenotypes: Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__, for example,
           `Pheno <http://fastlmm.github.io/PySnpTools/#snpreader-pheno>`__, or `SnpData <http://fastlmm.github.io/PySnpTools/#snpreader-snpdata>`__.
           If you give a string, it should be the name of a `Pheno <http://fastlmm.github.io/PySnpTools/#snpreader-pheno>`__-formatted file.
           If only one phenotype is given, any individual with missing phenotype data will be removed from processing.
           If multiple phenotypes are given, any missing data will raise an error. (Avoid this by removing missing values, perhaps via filling in
           missing values.)
    :type pheno: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__ or a string

    :param G0: SNPs from which to create a similarity matrix. Defaults to ``test_snps``.
           Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__.
           If you give a string, it should be the name of a `Bed <http://fastlmm.github.io/PySnpTools/#snpreader-bed>`__-formatted file.
           For cluster runs, this and ``test_snps`` should be readable from any node on the cluster, for example, by using
           `DistributedBed <http://fastlmm.github.io/PySnpTools/#snpreader-distributedbed>`__ format.
    :type G0: `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__ or a string

    :param covar: covariate information, optional: Can be any `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__, for example, `Pheno <http://fastlmm.github.io/PySnpTools/#snpreader-pheno>`__,
           or `SnpData <http://fastlmm.github.io/PySnpTools/#snpreader-snpdata>`__,.
           If you give a string, it should be the name of a `Pheno <http://fastlmm.github.io/PySnpTools/#snpreader-pheno>`__-formatted file. #!!!LATER raise error if has NaN
    :type covar: a `SnpReader <http://fastlmm.github.io/PySnpTools/#snpreader-snpreader>`__ or a string

    :param cache: Tells where to store intermediate results. Place the cache on an SSD drive for best performance.
                  By default, the cache will be an automatically-erasing temporary directory. (If the TEMP environment variable is set,
                  Python places the temporary directory under it.)
                  A string can be given and will be interpreted as the path of a local directory to use as a cache. (The local
                  directory will **not** be automatically erased and so must be user managed.) A `FileCache <http://fastlmm.github.io/PySnpTools/#util-filecache-filecache>`__
                  instance can be given, which provides a
                  method to specify a cluster-distributed cache. (`FileCache <http://fastlmm.github.io/PySnpTools/#util-filecache-filecache>`__'s will **not** be automatically erased and must be user managed.)
                  Finally, a dictionary from 0 to 22 (inclusive) to `FileCache <http://fastlmm.github.io/PySnpTools/#util-filecache-filecache>`__ (or None or string)
                  can be given. The dictionary specifies a cache for general work (0) and for every chromosome (1 to 22, inclusive).
    :type cache: None or string or `FileCache <http://fastlmm.github.io/PySnpTools/#util-filecache-filecache>`__ or dictionary from number to a `FileCache <http://fastlmm.github.io/PySnpTools/#util-filecache-filecache>`__.

    :param memory_factor: How much memory to use proportional to ``G0``, optional.
            If not given, will assume that it can use memory about the same size as one copy of ``G0``.
    :type memory_factor: number

    :param output_file_name: Name of file to write results to, optional. If not given, no output file will be created. The output format is tab-delimited text.
    :type output_file_name: file name

    :param runner: a `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__, optional: Tells how to run locally, multi-processor, or on a cluster.
        If not given, the function is run locally.
    :type runner: `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__

    :param min_work_count: When running the work on a cluster, the **minimum** number of pieces in which to divide the work. Defaults to 1, which is usually fine.
    :type min_work_count: integer

    :param gtg_runner: the `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__ to use instead of ``runner`` for the GtG stage of work.
        For an overview of the stages, see below.
    :type gtg_runner: `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__

    :param gtg_min_work_count: the min_work_count to use instead of ``min_work_count`` on the GtG stage of work.
    :type gtg_min_work_count: integer

    :param svd_runner: the `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__ to use instead of ``runner`` for the SVD stage of work.
    :type svd_runner: `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__

    :param postsvd_runner: the `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__ to use instead of ``runner`` for the PostSVD stage of work.
    :type postsvd_runner: `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__.

    :param postsvd_min_work_count: the min_work_count to use instead of ``min_work_count`` on the PostSVD stage of work.
    :type postsvd_min_work_count: integer

    :param test_snps_runner: the `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__ to use instead of ``runner`` for the TestSNPS stage of work.
    :type test_snps_runner: a `Runner <http://fastlmm.github.io/PySnpTools/#util-mapreduce1-runner-runner>`__.

    :param test_snps_min_work_count: the min_work_count to use instead of ``min_work_count`` on the TestSNPS stage of work.
    :type test_snps_min_work_count: integer

    :param count_A1: If it needs to read SNP data from a BED-formatted file, tells if it should count the number of A1
         alleles (the PLINK standard) or the number of A2 alleles. False is the current default, but in the future the default will change to True.
    :type count_A1: bool

    :param clear_local_lambda: A function to run in the middle of the PostSVD stage, typically to clear unneeded large files from the local file cache.
    :type clear_local_lambda: function or lambda

    :param force_python_only: (Default: False) Skip faster C++ code. Used for debugging and testing.

    :rtype: Pandas dataframe with one row per test SNP. Columns include "PValue"

    :Example:

    >>> import logging
    >>> from fastlmm.association import single_snp
    >>> from pysnptools.snpreader import Bed
    >>> from fastlmm.util import example_file # Download and return local file name
    >>> logging.basicConfig(level=logging.INFO)
    >>> test_snps = Bed(example_file('tests/datasets/synth/all.*','*.bed'),count_A1=True)[:,::10] #use every 10th SNP
    >>> pheno_fn = example_file("tests/datasets/synth/pheno_10_causals.txt")
    >>> cov_fn = example_file("tests/datasets/synth/cov.txt")
    >>> results_dataframe = single_snp_scale(test_snps=test_snps, pheno=pheno_fn, covar=cov_fn, count_A1=False)
    -etc-
    >>> print(results_dataframe.iloc[0].SNP,round(results_dataframe.iloc[0].PValue,7),len(results_dataframe))
    snp1200_m0_.37m1_.36 0.0 500

    The stages of processing are:

    * 0: G - Read G0 (the selected SNPs), standardize, regress out covariates, output G
    * 1: GtG - Compute GtG = G.T x G
    * 2: SVD - For each chromosome in the TestSNPs, extract the GtG for the chromosome, compute the singular value decomposition (SVD) on that GtG
    * 3: PostSVD - Transform the 22 GtG SVD into 22 G SVD results called U1 .. U22. Find h2, the importance of person-to-person similarity.
    * 4: TestSNPs - For each test SNP, read its data, regress out covariates, use the appropriate U and compute a Pvalue.

    All stages cache intermediate results. If the results for stage are found in the cache, that stage will be skipped.
    """
    with patch.dict("os.environ", {"ARRAY_MODULE": "numpy"}) as _:
        # Fill in with the default runner and default min_work_count
        gtg_runner = gtg_runner or runner
        gtg_min_work_count = gtg_min_work_count or min_work_count
        svd_runner = svd_runner or runner
        postsvd_runner = postsvd_runner or runner
        postsvd_min_work_count = postsvd_min_work_count or min_work_count
        test_snps_runner = test_snps_runner or runner
        test_snps_min_work_count = test_snps_min_work_count or min_work_count

        #######################################
        # all in time and space 1M x 3
        #######################################
        G0 = G0 or K0 or test_snps
        chrom_list, pheno1, RxY, test_snps1, X, Xdagger, G0 = preload(
            covar, G0, pheno, test_snps, count_A1=count_A1, multi_pheno_is_ok=True
        )
        cache_dict = _cache_dict_fixup(cache, chrom_list)

        G0_memmap_lambda, ss_per_snp = get_G0_memmap(
            G0, cache_dict[0], X, Xdagger, memory_factor
        )

        gtg_npz_lambda = get_gtg(
            cache_dict[0],
            G0.iid_count,
            G0.sid,
            G0_memmap_lambda,
            memory_factor,
            gtg_runner,
            min_work_count=gtg_min_work_count,
        )

        svd(
            chrom_list,
            gtg_npz_lambda,
            memory_factor,
            cache_dict[0],
            G0.iid_count,
            G0.pos,
            ss_per_snp,
            X,
            svd_runner,
        )

        log_frequency = 200 if logging.getLogger().level <= logging.INFO else 0
        postsvd(
            chrom_list,
            gtg_npz_lambda,
            memory_factor,
            cache_dict,
            G0.iid,
            G0.sid,
            G0_memmap_lambda,
            ss_per_snp,
            RxY,
            X,
            postsvd_runner,
            clear_local_lambda,
            postsvd_min_work_count,
            log_frequency=log_frequency,
        )

        test_snps_memory_factor = memory_factor

        frame = do_test_snps(
            cache_dict,
            chrom_list,
            gtg_npz_lambda,
            test_snps_memory_factor,
            G0.iid_count,
            G0.sid_count,
            G0.pos,
            pheno1,
            ss_per_snp,
            RxY,
            X,
            Xdagger,
            test_snps=test_snps1,
            runner=test_snps_runner,
            output_file_name=output_file_name,
            min_work_count=test_snps_min_work_count,
        )

        return frame


def snp_fn(data_folder, file_index):
    return data_folder + "/{0}.bed".format(file_index)


def pheno_fn(data_folder, file_index):
    return data_folder + "/pheno.{0}.txt".format(file_index)


def covar_fn(data_folder, file_index):
    return data_folder + "/covar.{0}.txt".format(file_index)


def generate(data_folder, iid_count, file_count, sid_count, seed=0):
    from pysnptools.util import snp_gen
    from pysnptools.util.generate import _generate_phenotype

    for file_index in range(file_count):
        snp_filename = snp_fn(data_folder, file_index)
        pheno_filename = pheno_fn(data_folder, file_index)
        covar_filename = covar_fn(data_folder, file_index)

        if file_index > 0:
            iid = [
                [iid[0], "{0}.{1}".format(iid[1], file_index)]
                for iid in full_snp_data.iid
            ]
            if not os.path.exists(snp_filename):
                snp_data1 = SnpData(
                    iid=iid,
                    sid=full_snp_data.sid,
                    val=full_snp_data.val,
                    pos=full_snp_data.pos,
                )
                Bed.write(snp_filename, snp_data1)
            if not os.path.exists(pheno_filename):
                pheno_data1 = SnpData(iid=iid, sid=phenodata.sid, val=phenodata.val)
                Pheno.write(pheno_filename, pheno_data1)
            if not os.path.exists(covar_filename):
                covar_data1 = SnpData(iid=iid, sid=covardata.sid, val=covardata.val)
                Pheno.write(covar_filename, covar_data1)

        else:
            #########################
            # SNPs
            #########################
            if not os.path.exists(snp_filename):
                logging.info("Creating '{0}'".format(snp_filename))
                # for 1000 x 4M takes about 6 clock minutes and 36 gig of memory
                full_snp_data = snp_gen(
                    fst=0.1,
                    dfr=0,
                    iid_count=iid_count // file_count,
                    sid_count=sid_count,
                    chr_count=22,
                    label_with_pop=True,
                    seed=seed,
                )
                # writing 1000 x 4M takes about 1.5 clock minutes (1 proc), 1 gig on
                Bed.write(snp_filename, full_snp_data)
                some_snp_data = full_snp_data[:, ::100]
            else:
                full_snp_data = Bed(snp_filename).read()
                some_snp_data = Bed(snp_filename)[
                    :, ::100
                ].read()  # Reading takes 5 clock secs (1 proc) and .5 gig of memory

            #########################
            # Pheno
            #########################
            if not os.path.exists(pheno_filename):
                logging.info("Creating '{0}'".format(pheno_filename))
                phenodata = SnpData(
                    iid=some_snp_data.iid,
                    sid=["pheno"],
                    val=_generate_phenotype(
                        some_snp_data, 10, genetic_var=0.5, noise_var=0.5, seed=seed
                    ).reshape(-1, 1),
                )
                Pheno.write(pheno_filename, phenodata)
            else:
                phenodata = Pheno(pheno_filename).read()

            #########################
            # Covar
            #########################
            if not os.path.exists(covar_filename):
                logging.info("Creating '{0}'".format(covar_filename))
                np.random.seed(seed)
                cov_val_0 = (
                    phenodata.val[:, 0] * 0.0001
                    + np.random.normal(size=some_snp_data.iid_count) * 1
                )
                cov_val_1 = np.random.normal(size=some_snp_data.iid_count) * 2
                covardata = SnpData(
                    iid=some_snp_data.iid,
                    sid=["covar0", "covar2"],
                    val=np.c_[cov_val_0, cov_val_1],
                )
                Pheno.write(covar_filename, covardata)
            else:
                covardata = Pheno(covar_filename).read()


def read_pheno_cache(data_folder, file_index_end, prefix, fn_fp):
    merge_fn = data_folder + "/{0}.0.{1}.merge.npz".format(prefix, file_index_end)
    if not os.path.exists(merge_fn):
        snpreader = _MergeIIDs(
            [
                Pheno(fn_fp(data_folder, file_index))
                for file_index in range(file_index_end)
            ]
        ).read()
        SnpNpz.write(merge_fn, snpreader)
    else:
        snpreader = SnpNpz(merge_fn)
    return snpreader


def bed_list(data_folder, snp0, file_index_end, iid_list_or_none=None):
    return [
        Bed(
            snp_fn(data_folder, file_index),
            sid=snp0.sid,
            pos=snp0.pos,
            iid=(None if iid_list_or_none is None else iid_list_or_none[file_index]),
            skip_format_check=True,
        )
        for file_index in range(file_index_end)
    ]


def read_bed_cache(data_folder, file_index_end):
    snp0 = Bed(snp_fn(data_folder, 0), skip_format_check=True)
    snps_iid_list_fn = data_folder + "/snps.0.{0}.iid_lists.npz".format(file_index_end)
    if os.path.exists(snps_iid_list_fn):
        with np.load(snps_iid_list_fn) as data:
            iid_list = data["iid_list"]
    else:
        iid_list = [
            bed.iid for bed in bed_list(data_folder, snp0, file_index_end, None)
        ]
        np.savez(snps_iid_list_fn, iid_list=iid_list)
    snps_merge_fn = data_folder + "/snps.0.{0}.mergeiids.npz".format(file_index_end)
    snp = _MergeIIDs(
        bed_list(data_folder, snp0, file_index_end, iid_list),
        cache_file=snps_merge_fn,
        skip_check=True,
    )
    return snp


# def scale_test():
#    import numpy as np
#    import numpy.linalg as la
#    import time
#    for sid_count in xrange(10000,50000,2500):
#        #iid_count = sid_count * 1000000 / 50000
#        iid_count = sid_count
#        x = np.random.randint(3, size=(sid_count,iid_count)).T
#        #print "creating ata"
#        #t = time.time()
#        #ata = np.dot(x.T,x)
#        #diff0 = time.time()-t
#        #print "ata ready"
#        t = time.time()
#        [a,b,c] = la.svd(x,False,True) #!!!use big_svd?
#        diff = time.time()-t
#        print iid_count,sid_count,diff#0,diff


#!!! exactly the same as in single_snp.py
def _create_dataframe(row_count):
    dataframe = pd.DataFrame(
        index=np.arange(row_count),
        columns=(
            "sid_index",
            "SNP",
            "Chr",
            "GenDist",
            "ChrPos",
            "PValue",
            "SnpWeight",
            "SnpWeightSE",
            "Mixing",
            "Nullh2",
        ),
    )
    #!!Is this the only way to set types in a dataframe?
    dataframe["sid_index"] = dataframe["sid_index"].astype(float)
    dataframe["Chr"] = dataframe["Chr"].astype(float)
    dataframe["GenDist"] = dataframe["GenDist"].astype(float)
    dataframe["ChrPos"] = dataframe["ChrPos"].astype(float)
    dataframe["PValue"] = dataframe["PValue"].astype(float)
    dataframe["SnpWeight"] = dataframe["SnpWeight"].astype(float)
    dataframe["SnpWeightSE"] = dataframe["SnpWeightSE"].astype(float)
    # Not currently calculated: dataframe['EffectSize'] = dataframe['EffectSize'].astype(float)
    dataframe["Mixing"] = dataframe["Mixing"].astype(float)
    dataframe["Nullh2"] = dataframe["Nullh2"].astype(float)

    return dataframe


def compute_stats(
    start,
    stop,
    snpsKY,
    snpsKsnps,
    YKY,
    N,
    logdetK,
    iid_count,
    sid,
    pos,
    h2,
    covar_bias_count,
    pheno_or_none,
):
    """
    Computes the stats for the test SNPs. Contains no slow matrix work. All work is order #test SNPs
    """
    # ==============================================================
    # 4M -> 4M
    #    (all the final operations)
    # ==============================================================
    beta = snpsKY / snpsKsnps
    if np.isnan(beta.min()):
        logging.warning("NaN beta value seen, may be due to an SNC (a constant SNP)")
        beta[snpsKY == 0] = 0.0

    variance_explained_beta = snpsKY * beta
    r2 = YKY[np.newaxis, :] - variance_explained_beta
    variance_beta = r2 / (N - 1) / snpsKsnps
    sigma2 = r2 / N
    nLL = 0.5 * (logdetK + N * (np.log(2.0 * np.pi * sigma2) + 1))
    chi2stats = beta * beta / variance_beta
    p_values = stats.f.sf(chi2stats, 1, iid_count - (covar_bias_count + 1))[
        :, 0
    ]  # note that G.shape is the number of individuals

    dataframe = _create_dataframe(len(sid))
    if pheno_or_none is not None:
        dataframe["Pheno"] = pheno_or_none
    dataframe["sid_index"] = np.arange(start, stop)
    dataframe["SNP"] = np.array(
        sid, "str"
    )  # This will be ascii on Python2 and unicode on Python3
    dataframe["Chr"] = pos[:, 0]
    dataframe["GenDist"] = pos[:, 1]
    dataframe["ChrPos"] = pos[:, 2]
    dataframe["PValue"] = p_values
    dataframe["SnpWeight"] = beta[:, 0]
    dataframe["SnpWeightSE"] = np.sqrt(variance_beta[:, 0])
    # Not currently calculated: dataframe['EffectSize'] = None
    dataframe["Mixing"] = np.zeros((len(sid))) + 0
    dataframe["Nullh2"] = np.zeros((len(sid))) + h2
    return dataframe


def _internal_single2(G0_chrom, test_snps_chrom, pheno, covar):
    return frame


def preload(covar, G0, pheno, test_snps, count_A1, multi_pheno_is_ok=False):
    """
    Load 'covar' and 'pheno' into memory. Remove and reorder iids as needed from all inputs.
    Compute some other values based on 'covar' and 'pheno'
    """
    test_snps = _snps_fixup(test_snps, count_A1=count_A1)
    G0 = _snps_fixup(G0, count_A1=count_A1)
    pheno = _pheno_fixup(pheno, count_A1=count_A1).read()

    if not multi_pheno_is_ok:
        assert pheno.sid_count == 1, "Expect pheno to be just one variable"

    if pheno.sid_count == 1:
        pheno = pheno[(pheno.val == pheno.val)[:, 0], :]
    else:
        assert not np.isnan(
            pheno.val
        ).any(), (
            "When multiple phenotypes are given, they must not have any missing values."
        )

    covar = _pheno_fixup(covar, iid_if_none=pheno.iid, count_A1=count_A1)
    chrom_list = list(
        set(test_snps.pos[:, 0])
    )  # find the set of all chroms mentioned in test_snps, the main testing data
    G0, test_snps, pheno, covar = pstutil.intersect_apply(
        [G0, test_snps, pheno, covar]
    )  #!!!is this time well spent?

    X = np.c_[
        covar.read(view_ok=True, order="A").val, np.ones((covar.iid_count, 1))
    ]  # view_ok because np.c_ will allocation new memory
    y = pheno.read(
        view_ok=True, order="A"
    ).val  # view_ok because this code already did a fresh read to look for any missing values

    # =============================================================
    # 1M x 3 -> 3 x 1M
    #    Only need to do this once, not 22 times
    # =============================================================
    Xdagger = la.pinv(X)  # SVD-based, and seems fast
    # =============================================================
    # 3 x 1M x 2 => 3 x 2
    #  Move out of chrom loop
    # =============================================================
    beta_y = Xdagger.dot(y)
    # =============================================================
    # 1M x 3 x 2 => 1M x 2
    #  Move out of chrom loop
    # =============================================================
    RxY = y - X.dot(beta_y)
    return chrom_list, pheno, RxY, test_snps, X, Xdagger, G0


def get_clear_local(file_cache):
    def sub_dir_path(d):
        return list(filter(os.path.isdir, [os.path.join(d, f) for f in os.listdir(d)]))

    def clear_local_lambda():
        for local_root in set(
            storage.local_lambda()[1] for storage in file_cache.values()
        ):
            for sub in sub_dir_path(local_root):
                logging.info("Removing '{0}'".format(sub))
                shutil.rmtree(sub)

    return clear_local_lambda


def get_G0_memmap(G0, file_cache_parent, X, Xdagger, memory_factor):
    """
    Create a version of the selected SNPs that is unit standardized. It will also be 'DiagKToN' standardization ready (see PySnpTools for info on 'DiagKToN'). The result
    will be uploaded into a file_cache (e.g. some shared cluster storage). The actual return value from this function is a lambda that when called will download G0 (if needed) and open it
    up as a memory-mapped numpy array.
    """
    file_cache = file_cache_parent.join("0_G")
    fn_done = "done.txt"
    fn_G0_memmap = "G0_data.memmap"
    fn_ss = "ss_per_snp.npz"

    def G0_memmap_lambda():
        with file_cache.open_read(fn_G0_memmap) as local_G0_memmap:
            G0_memmap = SnpMemMap(local_G0_memmap)
        #!!!not fully closing the read here because SnpMemMap has the local file open, but at least we know usage has started
        return G0_memmap

    if file_cache.file_exists(fn_done):
        with file_cache.open_read(fn_ss) as ss_storage:
            with np.load(ss_storage) as data:
                ss_per_snp = data["arr_0"]
        return G0_memmap_lambda, ss_per_snp

    file_cache.rmtree("")

    assert file_cache.file_exists(fn_G0_memmap) == file_cache.file_exists(
        fn_ss
    ), "expect '{0}' and '{1}' to either both exist or neither".format(
        fn_G0_memmap, fn_ss
    )

    work_count2 = -(
        G0.iid_count // -int(G0.sid_count * memory_factor)
    )  # Find the work count based on batch size (rounding up)
    work_count2 = max(work_count2, 2)  # Just to test it

    def debatch_closure2(work_index2):
        return G0.sid_count * work_index2 // work_count2

    logging.info("About to allocate memmap of G0_data.memmap")

    with _file_transfer_reporter(
        "G0_data.memmap upload", size=0, updater=None
    ) as updater2:
        with file_cache.open_write(
            fn_G0_memmap, size=8 * G0.iid_count * G0.sid_count, updater=updater2
        ) as G0_memmap_storage_file_name:
            G0_data_memmap = SnpMemMap.empty(
                iid=G0.iid,
                sid=G0.sid,
                filename=G0_memmap_storage_file_name,
                pos=G0.pos,
                order="F",
            )  #!!!is this the best order? dtype default ok?
            logging.info("Finished with allocation of memmap of G0_data.memmap")
            ss_per_snp = np.empty([G0.sid_count])
            t0 = time.time()
            for work_index in range(work_count2):
                if work_count2 > 1:
                    logging.info(
                        "get_G0_data_memmap: Working on part {0} of {1}".format(
                            work_index, work_count2
                        )
                    )
                start = debatch_closure2(work_index)
                stop = debatch_closure2(work_index + 1)

                ##############################################################
                ############# BIG ############################################
                ##############################################################
                # 1M x 25K -> 1M x 25K
                # Read all SNPs and Unit standardize
                #    Can be clustered by SNP, although output is 200 GB
                #            (Just remember unit trained and apply later?
                #   Use memory mapped files? SPARK? would be nice if could run on carlk4
                # ===========================================================
                G0_data_piece = G0[:, start:stop].read().standardize(Unit())
                ss_per_snp[start:stop] = np.einsum(
                    "i...,i...", G0_data_piece.val, G0_data_piece.val
                )  # Find the (G0_data.val**2).sum(0) without memory allocation
                #!!!Might be slightly better if this code and the original FaSTLMM code standardized the diagonal AFTER regressing away the covariants

                ##=============================================================
                ## 3 x 1M x 25K -> 3 x 25K
                ##=============================================================
                beta_G = Xdagger.dot(G0_data_piece.val)

                ##=============================================================
                # 1M x 3 x 25K -> 3 x 25K
                #   should be able to do this in place
                # =============================================================
                G0_data_piece.val -= X.dot(beta_G)  # mem x 2
                G0_data_memmap.val[:, start:stop] = G0_data_piece.val
            t1 = time.time()
            logging.info("G0 work took {0}".format(format_delta(t1 - t0)))
            logging.info("About to get name of file to np.savez ss_per_snp")
            with file_cache.open_write(fn_ss) as ss_storage:
                logging.info("About to np.savez ss_per_snp in '{0}'".format(ss_storage))
                np.savez(ss_storage, ss_per_snp)
                logging.info("About to close ss_per_snp")
            t2 = time.time()
            logging.info("done with ss_per_snp in {0}".format(format_delta(t2 - t1)))

            logging.info(
                "About to SnpMemMap.close G0_data_memmap in '{0}'".format(
                    G0_memmap_storage_file_name
                )
            )

            G0_data_memmap.flush()
            t3 = time.time()
            logging.info(
                "done with SnpMemMap.close G0_data_memmap in {0}".format(
                    format_delta(t3 - t2)
                )
            )
            logging.info("About to close g0.npz and g0.memmap")

    file_cache.save(fn_done, "")
    t4 = time.time()
    logging.info("Done with g0.npz and g0.memmap in {0}".format(format_delta(t4 - t3)))

    return G0_memmap_lambda, ss_per_snp


def get_gtg(
    common_cache_parent,
    G0_iid_count,
    G0_sid,
    G0_memmap_lambda,
    memory_factor,
    runner,
    min_work_count=1,
    force_python_only=False,
):
    """
    Compute G0.T x G0. Do the calculation in pieces with the smallest atom of work being done in multithreaded C++ code streaming the input matrix values from an file on (SSD) disk.
    The product, GtG will be stored in some cluster storage location.
    The actual return value of this function is a lambda. When that lambda is called, it retrieves GtG from storage and returns a local copy.
    """
    common_cache = common_cache_parent.join("1_GtG")
    fn_done = "done.txt"
    fn_gtg = "gtg.npz"

    def reader_closure():
        with common_cache.open_read(fn_gtg) as local_fn_gtg:
            gtg_npz = KernelNpz(local_fn_gtg)
        return gtg_npz

    if common_cache.file_exists(fn_done):
        return reader_closure

    common_cache.rmtree("")
    G0_sid_count = len(G0_sid)

    def writer_closure(gtg_data):
        with common_cache.open_write(fn_gtg, gtg_data.iid_count**2 * 8) as local_gtg:
            gtg_npz = KernelNpz.write(local_gtg, gtg_data)
        return gtg_npz

    ##############################################################
    ############# SLOWEST#########################################
    ##############################################################
    # 25K x 1M x 25K -> 25K x 25K
    # =============================================================

    logging.info("Computing G0.T x G0")
    work_count = -(
        G0_iid_count // -int(G0_sid_count * memory_factor)
    )  # Find the work count based on batch size (rounding up)
    work_count = max(
        min_work_count, work_count
    )  # Always divide the work into at least min_work_count parts
    work_count = min(
        work_count, len(G0_sid)
    )  # Don't divide the work more than the number of test SNPs

    t0_gtg = time.time()
    name = "{0}.get_gtg".format(
        os.path.basename(common_cache.name)
    )  #!!! little bug: if tree_cache.name ends with "/" or "\" doesn't work
    mmultfile_ata(
        G0_memmap_lambda,
        writer_closure,
        G0_sid,
        work_count,
        name,
        runner=runner,
        force_python_only=force_python_only,
    )
    common_cache.save(fn_done, "")
    logging.info("gtg: clocktime {0}".format(format_delta(time.time() - t0_gtg)))

    return reader_closure


def get_U(chrom, chrom_cache):
    chrom = int(chrom)

    fn_U = "3_PostSVD/chrom{0}/U.memmap".format(chrom)  #!!!const
    fn_UUYetc = "3_PostSVD/chrom{0}/UUYetc.npz".format(chrom)  #!!!const

    assert chrom_cache.file_exists(fn_U) == chrom_cache.file_exists(
        fn_UUYetc
    ), "expect '{0}' and '{1}' to either both exist or neither".format(fn_U, fn_UUYetc)
    assert chrom_cache.file_exists(fn_U), "expect '{0}'".format(fn_U)

    with chrom_cache.open_read(fn_UUYetc) as local_fn_UUYetc:
        with np.load(local_fn_UUYetc) as data:
            S = data["S"]
            UY = data["UY"]
            UUY = data["UUY"]

    with chrom_cache.open_read(fn_U) as local_fn_U:
        U_memmap = SnpMemMap(local_fn_U)

    return S, U_memmap, UY, UUY


def apply_h2(
    h2, S, UYUY, UUYUUYsum0, N, k
):  #!!!make more matrix like? (or multiproc???)
    result_list = []
    for pheno_index in range(len(UUYUUYsum0)):
        result = apply_h2_inner(
            h2[pheno_index], S, UYUY[:, [pheno_index]], UUYUUYsum0[[pheno_index]], N, k
        )
        result_list.append(result)
    logdetK = np.array([logdetK for logdetK, YKY, Sd, denom in result_list])
    YKY = np.array([YKY for logdetK, YKY, Sd, denom in result_list]).reshape(-1)
    Sd = np.array([Sd for logdetK, YKY, Sd, denom in result_list]).transpose()
    denom = np.array([denom for logdetK, YKY, Sd, denom in result_list])
    return logdetK, YKY, Sd, denom


def apply_h2_inner(h2, S, UYUY, UUYUUYsum0, N, k):
    Sd = h2 * S + (1.0 - h2)  # 25K
    denom = 1.0 - h2
    YKY = (UYUY / Sd.reshape(-1, 1)).sum(
        0
    ) + UUYUUYsum0 / denom  # Can be done in blocks
    logdetK = np.log(Sd).sum() + (N - k) * np.log(denom)
    return logdetK, YKY, Sd, denom


def find_h2(k, N, UUYUUYsum0, UYUY, S, chrom):
    h2_list = []
    for pheno_index in range(len(UUYUUYsum0)):
        h2_inner = find_h2_inner(
            k, N, UUYUUYsum0[[pheno_index]], UYUY[:, [pheno_index]], S, chrom
        )
        h2_list.append(h2_inner)
    return np.array(h2_list)


def find_h2_inner(k, N, UUYUUYsum0, UYUY, S, chrom):
    resmin = [None]

    def f(x, resmin=resmin):
        h2 = float(x)
        logdetK, YKY, _, _ = apply_h2_inner(h2, S, UYUY, UUYUUYsum0, N, k)
        sigma2 = YKY / N
        nLL = 0.5 * (logdetK + N * (np.log(2.0 * np.pi * sigma2) + 1))
        if (resmin[0] is None) or (nLL < resmin[0]["nLL"]):
            resmin[0] = {"nLL": nLL, "h2": h2}
        logging.debug("search\t{0}\t{1}".format(h2, nLL))
        return nLL

    # =============================================================
    # 25K x about 30
    # =============================================================
    _ = minimize1D(f=f, nGrid=10, minval=0, maxval=0.99999)
    h2 = resmin[0]["h2"]
    logging.info("h2={0}".format(h2))

    logdetK, YKY, Sd, denom = apply_h2_inner(h2, S, UYUY, UUYUUYsum0, N, k)
    return h2


def get_h2(k, N, UUYUUYsum0, UYUY, S, chrom_cache, chrom):
    fn_h2 = "3_PostSVD/chrom{0}/h2.npz".format(int(chrom))  #!!!const
    with chrom_cache.open_read(fn_h2) as local_fn_h2:
        with np.load(local_fn_h2) as data:
            h2 = data["arr_0"]
    logdetK, YKY, Sd, denom = apply_h2(h2, S, UYUY, UUYUUYsum0, N, k)
    return h2, logdetK, YKY, Sd, denom


def svd(
    chrom_list,
    gtg_npz_lambda,
    memory_factor,
    common_cache_parent,
    G0_iid_count,
    G0_pos,
    ss_per_snp,
    X,
    runner_svd,
):
    """
    For the chromosomes listed, compute an SVD on a square matrix SNP-to-SNP matrix. Each SVD can be done on a different
    node in a cluster. The actual SVD is done with special version of the LAPACK DGESDD function.

    Results are stored at a known location in the cluster storage.
    """

    common_cache = common_cache_parent.join("2_SVD")

    needed_chrom_list = [
        chrom
        for chrom in chrom_list
        if not common_cache.file_exists("done{0}.txt".format(int(chrom)))
    ]
    if not needed_chrom_list:
        return []

    def mapper_closure(chrom):
        logging.info("Caching chrom {0}".format(chrom))
        gtg_npz = gtg_npz_lambda()
        idx = G0_pos[:, 0] != chrom
        factor = float(G0_iid_count) / ss_per_snp[idx].sum()

        ##############################################################
        ############# SLOWEST ########################################
        ##############################################################
        # 1M x 25K x 25K => 1M x 25K
        # =============================================================
        idx2 = np.arange(len(idx))[idx]
        ata = gtg_npz[idx2].read()

        # Because the iid_count can be so big, the factor created for one_step_svd can be very inappropriate for here
        factor_tall_skinny = float(ata.iid_count) / np.diag(ata.val).sum()
        ata._val *= factor_tall_skinny

        ##############################################################
        ############# SLOWEST ########################################
        ##############################################################
        # 25K x 25K x 25K -> 25K x 25K
        # =============================================================
        num_threads = get_num_threads(None)
        logging.info(
            "About to svd on square {0}. Expected time ({2} procs)={1}".format(
                ata.iid_count,
                format_delta((ata.iid_count * 0.000707) ** 3 * 20.0 / num_threads),
                num_threads,
            )
        )
        t0 = time.time()
        [Uata3, Sata3, _] = la.svd(ata.val, full_matrices=False, compute_uv=True)
        logging.info(
            "Actual time for svd on square={0}".format(format_delta(time.time() - t0))
        )
        Sata3 *= factor / factor_tall_skinny  # make the results match one_step_svd
        S3 = Sata3**0.5
        V3 = Uata3.T

        ##############################################################
        ############# SLOWEST ########################################
        ##############################################################
        # 25K**2.8 -> 25K x 25K
        #        Is there a faster way to do the dot with the diag???
        # =============================================================
        SVinv3 = la.inv(np.dot(np.diag(S3), V3))
        SVinv3 *= np.sqrt(factor)

        S = S3
        if np.any(S < -0.1):
            logging.warning("kernel contains a negative Eigenvalue")
        inonzero = S > 1e-10
        S = S[inonzero]
        S = S * S

        SVinv3b = SVinv3[:, inonzero]

        fn = "SVinv_etc{0}.npz".format(int(chrom))  #!!!const
        if common_cache.file_exists(fn):
            common_cache.remove(fn)
        with common_cache.open_write(fn) as local_file_name:
            np.savez(local_file_name, SVinv3b=SVinv3b, S=S)
        common_cache.save("done{0}.txt".format(int(chrom)), "")
        return fn

    SVinv_etc_fn_list = map_reduce(
        needed_chrom_list,
        mapper=mapper_closure,
        runner=runner_svd,
        name="{0}.svd".format(os.path.basename(common_cache.name)),
        input_files=[],
        output_files=[],
    )
    return SVinv_etc_fn_list


def postsvd_piece(
    start_iid_index,
    stop_iid_index,
    G0_memmap,
    idx_array,
    SVinv3b,
    fn_U_piece,
    log_frequency=-1,
):
    logging.info("About to read")
    t0_piece = time.time()
    piece = np.zeros((stop_iid_index - start_iid_index, SVinv3b.shape[0]), order="F")
    with open(G0_memmap.filename, "rb") as fp:
        so_far_idx = 0
        for start_idx, stop_idx in idx_array:
            start_idx, stop_idx = int(start_idx), int(
                stop_idx
            )  # We convert from numpy.int32 to python ints so that their product will be a python long instead of a (overflowed) numpy.int32
            fp.seek(
                G0_memmap.offset + start_idx * G0_memmap.iid_count * 8
            )  # Skip sids that are not of interest
            for sid_index in range(so_far_idx, so_far_idx + stop_idx - start_idx):
                if log_frequency > 0 and sid_index % log_frequency == 0:
                    logging.info(
                        "on sid_index {0} of {1}".format(sid_index, SVinv3b.shape[0])
                    )
                row = np.fromfile(
                    fp, dtype=np.float64, count=G0_memmap.iid_count
                )  #!!! instead of reading whole column, how about just reading the piece of the column that is wanted? Whould this be a little faster?
                try:
                    piece[:, sid_index] = row[start_iid_index:stop_iid_index]
                except:
                    raise Exception(
                        "piece[:,{0}] = row[{1}:{2}]".format(
                            sid_index, start_iid_index, stop_iid_index
                        )
                    )
            so_far_idx += stop_idx - start_idx
    logging.info("About to mult")
    product = np.array(np.dot(piece, SVinv3b), order="F")
    logging.info("About to save")
    product.flatten(order="K").tofile(fn_U_piece, "")
    logging.info(
        "post svd piece: clocktime {0}".format(format_delta(time.time() - t0_piece))
    )


def postsvd(
    chrom_list,
    gtg_npz_lambda,
    memory_factor,
    cache_dict,
    G0_iid,
    G0_sid,
    G0_memmap_lambda,
    ss_per_snp,
    RxY,
    X,
    postsvd_runner,
    clear_local_lambda,
    min_work_count,
    log_frequency=-1,
):
    """
    For the chromosomes listed, take the square SVD and turn it into U, the "tall-and-skinny" SVD needed. Also, computed related values UY and UUY.
    Finally, find search for the best h2, which tells how much weight to give to person-to-person similarity vs. pure noise.

    This function uses two levels of map-reduce that are run as a single cluster job. The top level loops across the chromosomes, the
    second level does a matrix multiple in blocks. At the lowest level, the matrix multiple is multithreaded.

    Save the results under a known name in the cluster storage.
    """
    sub_dir = "3_PostSVD"
    common_cache = cache_dict[0].join(sub_dir)
    # cache_dict[int(chrom)].file_exists("h2_{0}.npz".format(int(chrom)))
    needed_chrom_list = [
        chrom
        for chrom in chrom_list
        if not common_cache.file_exists("chrom{0}/done.txt".format(int(chrom)))
    ]
    if not needed_chrom_list:
        return
    G0_iid_count = len(G0_iid)
    G0_sid_count = len(G0_sid)

    def mapper_closure(chrom):
        chrom = int(chrom)
        chrom_storage = cache_dict[chrom].join(sub_dir)

        fn_U = "chrom{0}/U.memmap".format(int(chrom))
        if chrom_storage.file_exists(fn_U):
            chrom_storage.remove(fn_U)
        fn_UUYetc = "chrom{0}/UUYetc.npz".format(int(chrom))
        if chrom_storage.file_exists(fn_UUYetc):
            chrom_storage.remove(fn_UUYetc)
        fn_h2 = "chrom{0}/h2.npz".format(int(chrom))  #!!!const
        if chrom_storage.file_exists(fn_h2):
            chrom_storage.remove(fn_h2)

        work_count = -(
            G0_iid_count // -int(G0_sid_count * memory_factor)
        )  # Find the work count based on batch size (rounding up)
        work_count = max(work_count, min_work_count)

        def mapper_closure_inner(work_index):
            SVinv_etc_fn = "2_SVD/SVinv_etc{0}.npz".format(chrom)
            with cache_dict[0].open_read(SVinv_etc_fn) as handle_local:
                with np.load(handle_local) as data:
                    SVinv3b = data["SVinv3b"]
                    S = data["S"]

            G0_memmap = G0_memmap_lambda()
            idx = G0_memmap.pos[:, 0] != chrom
            ##############################################################
            ############# SLOWEST ########################################
            ##############################################################
            # 1M x 25K x 25K => 1M x 25K

            sid = ["sid{0}".format(i) for i in range(SVinv3b.shape[1])]
            idx_intrangeset = IntRangeSet(
                i for i, keep in enumerate(idx) if keep
            )  # IntRangeSet('285:1000')
            idx_array = np.array(list(idx_intrangeset.ranges()))

            def debatch_closure(work_index):
                return G0_memmap.iid_count * work_index // work_count

            if work_count > 1:
                logging.info(
                    "post svd, chrom {0}: Working on part {1} of {2}".format(
                        chrom, work_index, work_count
                    )
                )

            start_iid_index = debatch_closure(work_index)
            stop_iid_index = debatch_closure(work_index + 1)

            fn_U_done = "chrom{0}/U_pieces/done{1}_{2}.txt".format(
                chrom, work_index, work_count
            )
            fn_U_piece = "chrom{0}/U_pieces/{1}_{2}.memmap".format(
                chrom, work_index, work_count
            )
            chrom_cache = cache_dict[chrom].join(sub_dir)
            if chrom_cache.file_exists(fn_U_done):
                logging.info(
                    "Piece '{0}' already exists, so skipping work".format(fn_U_piece)
                )
            else:
                if chrom_cache.file_exists(fn_U_piece):
                    chrom_cache.remove(fn_U_piece)
                with chrom_cache.open_write(fn_U_piece) as local_file_name:
                    postsvd_piece(
                        start_iid_index,
                        stop_iid_index,
                        G0_memmap,
                        idx_array,
                        SVinv3b,
                        local_file_name,
                        log_frequency=log_frequency,
                    )
                chrom_cache.save(fn_U_done, "")
            return (
                fn_U_piece,
                start_iid_index,
                stop_iid_index,
                len(sid) if work_index == 0 else None,
                S if work_index == 0 else None,
            )

        def reducer_closure_inner(fn_U_piece_sequence):
            if clear_local_lambda is not None:
                logging.info(
                    "To save space, removing all cached files from local drive"
                )
                clear_local_lambda()

            t0_start = time.time()
            logging.info(
                "Starting postsvd reduce for chrom '{0}' with file downloads".format(
                    chrom
                )
            )

            fn_U_piece_list = list(
                fn_U_piece_sequence
            )  # make it a list instead of generator, so can run through it twice
            _, _, _, sid_count, S = fn_U_piece_list[
                0
            ]  # Get sid_count and S from the first work item. We pass these this way to avoid bring in files to the local machine (if run with "map_reduceX") or to the reduce

            for fn_U_piece, start_iid_index, stop_iid_index, _, _ in fn_U_piece_list:
                with chrom_storage.open_read(
                    fn_U_piece
                ) as local_file_name:  #!!! is assuming that a local file with the name is OK.
                    pass

            t0_download = time.time()
            logging.info(
                "File downloads took {0}. Next putting U together".format(
                    format_delta(t0_download - t0_start)
                )
            )

            with chrom_storage.open_write(fn_U) as handle_fn_U_file_name:
                sid = ["sid{0}".format(i) for i in range(sid_count)]
                U_snp_mem_map = SnpMemMap.empty(
                    iid=G0_iid,
                    sid=sid,
                    filename=handle_fn_U_file_name,
                    dtype=np.float64,
                    order="F",
                )
                fp_list = (
                    []
                )  #!!! would be nice to have a try_catch to be sure all these get closed
                file_name_list = []
                for piece_index, (
                    fn_U_piece,
                    start_iid_index,
                    stop_iid_index,
                    _,
                    _,
                ) in enumerate(fn_U_piece_list):
                    with chrom_storage.open_read(
                        fn_U_piece
                    ) as handle_local:  # The local file is used after this is closed. Not nice, but OK in this case.
                        logging.info(
                            "About to open piece {0} of {1}".format(
                                piece_index, len(fn_U_piece_list)
                            )
                        )
                        fp_list.append(open(handle_local, "rb"))
                with open(U_snp_mem_map.filename, "r+b") as fp_U:
                    fp_U.seek(U_snp_mem_map.offset)
                    with log_in_place("Creating U file", logging.INFO) as log_writer:
                        for sid_index in range(sid_count):
                            if (
                                sid_index % 100 == 0
                            ):  #!!!use something like log_freq instead of '100'
                                log_writer(
                                    "On sid {0} of {1}".format(sid_index, sid_count)
                                )
                            prev_stop = 0
                            for piece_index, (
                                fn_U_piece,
                                start_iid_index,
                                stop_iid_index,
                                _,
                                _,
                            ) in enumerate(fn_U_piece_list):
                                assert start_iid_index == prev_stop, "real assert"
                                prev_stop = stop_iid_index
                                fp = fp_list[piece_index]
                                buf2 = np.fromfile(
                                    fp,
                                    dtype=np.float64,
                                    count=stop_iid_index - start_iid_index,
                                )
                                buf2.tofile(fp_U)
                            assert stop_iid_index == G0_iid_count, "real assert"
                for fp in fp_list:
                    fp.close()

                t0_U = time.time()
                logging.info(
                    "Putting U together took {0}. Next creating UY & UUY files".format(
                        format_delta(t0_U - t0_download)
                    )
                )

                # =============================================================
                # 25K x 1M x 2 -> 25K x 2
                # =============================================================
                UY = U_snp_mem_map.val.T.dot(
                    RxY
                )  # Note: This could be pushed into the 'map' step, with each step returning a UYi that would all be summed together.
                # =============================================================
                # 1M x 25K x 2 -> 1M x 2
                # =============================================================
                UUY = RxY - U_snp_mem_map.val.dot(
                    UY
                )  # This is can't be pushed into the 'map' step because it depends on all of UY

                U_snp_mem_map.flush()

            with chrom_storage.open_write(fn_UUYetc) as local_fn_UUYetc:
                np.savez(local_fn_UUYetc, S=S, UY=UY, UUY=UUY)

            # =============================================================
            # 25K x about 30
            # =============================================================
            UYUY = UY * UY
            UUYUUYsum0 = (UUY * UUY).sum(0)
            N = G0_iid_count - X.shape[1]  # number of degrees of freedom
            k = S.shape[0]
            h2 = find_h2(k, N, UUYUUYsum0, UYUY, S, chrom)  # h2 depends on y

            t0_etc = time.time()
            logging.info(
                "Creating related files took {0}. Next uploading files.".format(
                    format_delta(t0_etc - t0_U)
                )
            )

            with chrom_storage.open_write(fn_h2) as local_fn_h2:
                np.savez(local_fn_h2, h2)

            if clear_local_lambda is not None:
                chrom_storage.cloud_storage_only()

            common_cache.save("chrom{0}/done.txt".format(int(chrom)), "")

            t0_up = time.time()
            logging.info("Uploading took {0}.".format(format_delta(t0_up - t0_etc)))

            logging.info(
                "Postsvd Reduce Summary: Down {0}, Calc {1}, Up {2}, Total {3}".format(
                    format_delta(t0_download - t0_start),
                    format_delta(t0_etc - t0_download),
                    format_delta(t0_up - t0_etc),
                    format_delta(t0_up - t0_start),
                )
            )

        return map_reduce(
            range(work_count),
            mapper=mapper_closure_inner,
            reducer=reducer_closure_inner,
            name="{0}.postsvd_{1}".format(os.path.basename(chrom_storage.name), chrom),
            input_files=[],
            output_files=[],
        )

    map_reduceX(
        needed_chrom_list,
        mapper=mapper_closure,
        name="{0}.postsvd".format(os.path.basename(cache_dict[0].name)),
        runner=postsvd_runner,
    )


def get_U_h2(
    chrom,
    gtg_npz_lambda,
    memory_factor,
    chrom_cache,
    G0_iid_count,
    G0_pos,
    ss_per_snp,
    RxY,
    pheno,
    X,
):
    #!!! similar code above
    logging.debug("Retrieving on chrom {0}".format(chrom))
    idx = G0_pos[:, 0] != chrom
    factor = float(G0_iid_count) / ss_per_snp[idx].sum()

    ##############################################################
    ############# SLOWEST ########################################
    ##############################################################
    # 25K x 25K x 25K -> 25K x 25K
    # 1M x 25K x 25K => 1M x 25K
    # =============================================================
    S, U_memmap, UY, UUY = get_U(
        chrom, chrom_cache
    )  #!!!y: at the end, U is multiplied twice with yish things

    # =============================================================
    # 25K x about 30 x 3
    # =============================================================
    UYUY = UY * UY
    UUYUUYsum0 = (UUY * UUY).sum(0)
    N = pheno.iid_count - X.shape[1]  # number of degrees of freedom
    k = S.shape[0]
    h2, logdetK, YKY, Sd, denom = get_h2(
        k, N, UUYUUYsum0, UYUY, S, chrom_cache, chrom
    )  #!!!y: h2 depends on y

    return h2, U_memmap, Sd, denom, UY, UUY, YKY, N, logdetK


def __del__(test_snps):
    """
    Close any open Bed files
    """
    from pysnptools.snpreader._subset import _SnpSubset
    from pysnptools.snpreader import SnpData, _Distributed1Bed, DistributedBed
    from pysnptools.pstreader import _MergeRows, _MergeCols

    if isinstance(test_snps, Bed) or isinstance(test_snps, _Distributed1Bed):
        test_snps.__del__()
    elif isinstance(test_snps, SnpGen) or isinstance(test_snps, SnpData):
        pass
    elif isinstance(test_snps, DistributedBed):
        __del__(test_snps._merge)
    elif isinstance(test_snps, _SnpSubset):
        __del__(test_snps._internal)
    elif isinstance(test_snps, _MergeRows) or isinstance(test_snps, _MergeCols):
        for sub in test_snps.reader_list:
            __del__(sub)
    else:
        logging.warning("Don't know how to __del__ '{0}'".format(test_snps))


# code: no longer used (was for debugging???)
# def is_good(node,file,chrom):
#    #Only use the file if it 1. exists and 2. It's *.memmap is over 150000000000L in bytes in size
#    if not os.path.exists(file):
#        return False
#    big = pattern2.format(node,".U{0}.0.memmap".format(int(chrom)))
#    if not os.path.exists(big) or os.path.getsize(big) <  150000000000L:
#        return False
#    mid = pattern2.format(node,".U{0}.0.npz".format(int(chrom)))
#    if not os.path.exists(mid) or os.path.getsize(mid) <  5000000L:
#        return False
#    return True

# def random_source(chrom,pattern,node_list):
#    part = pattern.format(int(chrom))
#    shuffled_node_list = list(node_list[int(chrom)-1])
#    random.shuffle(shuffled_node_list)
#    for node in shuffled_node_list:
#        source = pattern2.format(node,part)
#        if is_good(node,source,chrom):
#            return source
#    raise Exception("Can't find '{0}' in {1}".format(part, node_list))


def do_test_snps(
    cache_dict,
    chrom_list,
    gtg_npz_lambda,
    memory_factor,
    G0_iid_count,
    G0_sid_count,
    G0_pos,
    pheno,
    ss_per_snp,
    RxY,
    X,
    Xdagger,
    test_snps,
    runner,
    output_file_name=None,
    min_work_count=1,
):
    """
    For every test SNP, measure its pvalue. Nest two levels of map-reduce. On the top level, loop over chromosomes. Within each chromosome
    loop over blocks of testSNPs. At the lowest-level, do the matrix-multiple (and related operators) with multithreaded C++.

    Return a Pandas frame of the final results.
    """
    __del__(test_snps)  # Close any open Bed files

    results_storage = cache_dict[0].join("4_TestSNPs")

    def mapper_closure2(chrom):
        chrom_index = chrom_list.index(chrom)
        chrom = int(chrom)
        chrom_storage = cache_dict[chrom]

        test_snps_chrom = test_snps[:, test_snps.pos[:, 0] == chrom]
        logging.info("test_snps has sid_count of {0}".format(test_snps.sid_count))
        logging.info(
            "test_snps_chrom({1}) has sid_count of {0}".format(
                test_snps_chrom.sid_count, chrom
            )
        )

        work_count = int(
            -(-test_snps_chrom.iid_count // (G0_sid_count * memory_factor))
        )  # Do blocks based on the size of G0
        work_count = max(
            min_work_count, work_count
        )  # Always divide the work into at least min_work_count parts
        work_count = min(
            work_count, test_snps_chrom.sid_count
        )  # Don't divide the work more than the number of test SNPs
        logging.info("work_count = {0}".format(work_count))

        def mapper_closure(work_index):
            done_file = "chrom{0}/done.{1}of{2}.txt".format(
                chrom, work_index, work_count
            )
            cache_file = "chrom{0}/result.{1}of{2}.tsv".format(
                chrom, work_index, work_count
            )
            if results_storage.file_exists(done_file):
                with results_storage.open_read(cache_file) as local_file:
                    dataframe = pd.read_csv(local_file, delimiter="\t")
                    return dataframe

            if results_storage.file_exists(cache_file):
                results_storage.remove(cache_file)

            def debatch_closure(work_index):
                start = test_snps_chrom.sid_count * work_index // work_count
                logging.debug(
                    "chrom={0},work_index={1},start={2},test_snps_chrom.sid_count={3}".format(
                        chrom, work_index, start, test_snps_chrom.sid_count
                    )
                )
                return start

            h2, U_memmap, Sd, denom, UY, UUY, YKY, N, logdetK = get_U_h2(
                chrom,
                gtg_npz_lambda,
                memory_factor,
                chrom_storage,
                G0_iid_count,
                G0_pos,
                ss_per_snp,
                RxY,
                pheno,
                X,
            )

            if work_count > 1:
                logging.info(
                    "single_low_snp: Working on snp block {0} of {1}".format(
                        work_index, work_count
                    )
                )
            start = debatch_closure(work_index)
            logging.debug(
                "A:chrom={0},work_index={1},start={2}".format(chrom, work_index, start)
            )
            stop = debatch_closure(work_index + 1)
            logging.debug(
                "A:chrom={0},work_index+1={1},end={2}".format(
                    chrom, work_index + 1, stop
                )
            )
            t0_gen = time.time()
            snps_read = test_snps_chrom[:, start:stop].read().standardize()
            logging.info(
                "test snp reader {0} of {1}, clocktime {2}".format(
                    work_index, work_count, format_delta(time.time() - t0_gen)
                )
            )
            logging.debug(
                "Xdagger {0}x{1}. snp_read {2}x{3}".format(
                    Xdagger.shape[0],
                    Xdagger.shape[1],
                    snps_read.iid_count,
                    snps_read.sid_count,
                )
            )

            # =============================================================
            #  Note: on all these the 4M is all the SNPs, but because we can cluster by chrom and blocks really 4M/22/10 (about 16K to 18K) per work item
            # 3 x 1M x 4M -> 3 x 4M
            # =============================================================
            logging.debug("beta_snps = Xdagger.dot(snps_read.val)")
            beta_snps = Xdagger.dot(snps_read.val)
            ##############################################################
            ########### BIG ##############################################
            ##############################################################
            # 1M x 3 x 4M -> 1M x 4M
            # =============================================================
            logging.debug("Rxsnps = snps_read.val - X.dot(beta_snps)")
            Rxsnps = snps_read.val - X.dot(beta_snps)
            ##############################################################
            ###########   SLOWEST ########################################
            ##############################################################
            # 25K X 1M x 4M -> 25K x 4M
            # =============================================================
            logging.debug("U_memmap '{0}'".format(U_memmap))
            logging.debug(
                "U_data {0}x{1}. Rxsnps {2}x{3}. time={4}".format(
                    U_memmap.iid_count,
                    U_memmap.sid_count,
                    Rxsnps.shape[0],
                    Rxsnps.shape[1],
                    datetime.now().strftime("%Y-%m-%d %H:%M"),
                )
            )
            logging.debug("Usnps = U_data.val.T.dot(Rxsnps)")

            ##############################################################
            ###########   SLOWEST ##### BIG ##############################
            ##############################################################
            # 25K x 1M x 4M -> 25K x 4M
            # and
            # 25K x 4M
            #          while at it, also (UUsnps * UUsnps).sum(0)
            # =============================================================
            # U_data = U_memmap.read(order='K',view_ok=True) #in the work loop so it can be run on cluster
            # UUsnps = Rxsnps - U_data.val.dot(Usnps)

            log_frequency = 10 if logging.getLogger().level <= logging.INFO else 0
            Usnps, UUsnps = mmultfile_b_less_aatb(
                U_memmap, Rxsnps, log_frequency=log_frequency
            )
            logging.debug(
                "U_data {0}x{1}. Usnps {2}x{3}. time={4}".format(
                    U_memmap.iid_count,
                    U_memmap.sid_count,
                    Usnps.shape[0],
                    Usnps.shape[1],
                    datetime.now().strftime("%Y-%m-%d %H:%M"),
                )
            )
            logging.debug("UUsnps = Rxsnps - U_data.val.dot(Usnps)")

            dataframe_list = []
            UUSnpsUUSnps_sum0 = (UUsnps * UUsnps).sum(0)
            for pheno_index in range(
                len(h2)
            ):  #!!!make more matrix like? (or multiprocess???)
                # ==============================================================
                # 25K x 4M -> 4M
                # ==============================================================
                logging.debug(
                    "time={0}".format(datetime.now().strftime("%Y-%m-%d %H:%M"))
                )
                logging.debug(
                    "snpsKsnps = ((Usnps / Sd.reshape(-1,1) * Usnps).sum(0) + (UUsnps * UUsnps).sum(0) / denom).reshape(-1,1) #Can be done in blocks"
                )
                snpsKsnps = (
                    (Usnps / Sd[:, [pheno_index]] * Usnps).sum(0)
                    + UUSnpsUUSnps_sum0 / denom[pheno_index]
                ).reshape(
                    -1, 1
                )  # Can be done in blocks
                # ==============================================================
                # 25K x 4M -> 25K x 4M
                # ==============================================================
                logging.debug(
                    "UAS = Usnps / np.lib.stride_tricks.as_strided(Sd, (Sd.size,Usnps.shape[1]), (Sd.itemsize,0))"
                )
                UAS = Usnps / np.lib.stride_tricks.as_strided(
                    Sd[:, [pheno_index]],
                    (Sd.shape[0], Usnps.shape[1]),
                    (Sd.itemsize, 0),
                )
                # ==============================================================
                # 4M x 25K x 1 + 4M x 1M x 1 => 4M
                # ==============================================================
                logging.debug(
                    "snpsKY = UAS.T.dot(UY) + UUsnps.T.dot(UUY) / denom #!!!y: some y work"
                )
                snpsKY = (
                    UAS.T.dot(UY[:, [pheno_index]])
                    + UUsnps.T.dot(UUY[:, [pheno_index]]) / denom[pheno_index]
                )
                # ==============================================================
                # 4M
                # ==============================================================
                assert test_snps_chrom.iid_count == U_memmap.iid_count
                pheno_or_none = None if pheno.sid_count == 1 else pheno.sid[pheno_index]
                dataframe_inner = compute_stats(
                    start,
                    stop,
                    snpsKY,
                    snpsKsnps,
                    YKY[[pheno_index]],
                    N,
                    logdetK[[pheno_index]],
                    snps_read.iid_count,
                    snps_read.sid,
                    snps_read.pos,
                    h2[pheno_index],
                    covar_bias_count=Xdagger.shape[0],
                    pheno_or_none=pheno_or_none,
                )
                dataframe_list.append(dataframe_inner)

            dataframe = pd.concat(dataframe_list)
            with results_storage.open_write(cache_file) as local_file:
                dataframe.to_csv(local_file, sep="\t", index=False)

            results_storage.save(done_file, "")

            return dataframe

        def reducer_closure(result_list):
            frame = pd.concat(result_list)
            frame.sort_values(by="PValue", inplace=True)
            frame.index = np.arange(len(frame))
            return frame

        frame = map_reduce(
            range(work_count),
            mapper=mapper_closure,
            name="{0}.test_snps_{1}".format(
                os.path.basename(cache_dict[0].name), chrom
            ),
            reducer=reducer_closure,
        )
        return frame

    def reducer_closure2(result_list2):
        frame = pd.concat(result_list2)
        frame.sort_values(by="PValue", inplace=True)
        frame.index = np.arange(len(frame))

        if output_file_name is not None:
            pstutil.create_directory_if_necessary(output_file_name)
            frame.to_csv(output_file_name, sep="\t", index=False)

        logging.info("PhenotypeName(s)\t{0}".format(pheno.sid))
        logging.info("SampleSize\t{0}".format(test_snps.iid_count))
        logging.info("SNPCount\t{0}".format(test_snps.sid_count))

        return frame

    name_prefix = "{0}.test_snps_".format(os.path.basename(cache_dict[0].name))
    frame = map_reduceX(
        chrom_list,
        mapper=mapper_closure2,
        reducer=reducer_closure2,
        name=lambda index: name_prefix
        + str(int(chrom_list[index])),  # Create a job name based on the chrom number
        runner=runner,
    )

    return frame


def clear_cache_dict(start_stage, cache_dict, chrom_num_list):
    """
    Remove selected intermediate results from the cache dictionary.

    :param start_stage: None (do nothing), 'all' (clear all), 'gtg' (clear stage gtg and later),
         'svd' (clear stage 'svd' and later), 'postsvd' (clear stage postsvd and later), 'testsnps' (clear stage testsnps)
    :type start_stage: string or None

    :param cache_dict: Dictionary mapping chromosomes to :class:`LocalCache` or other storage
    :type cache_dict: dictionary

    :param chrom_num_list: List of chromosome numbers to look at.
    :type chrom_num_list: list of numbers.

    """
    if start_stage is None:
        return

    with log_in_place(
        "clearing cache dict to '{0}'".format(start_stage), logging.INFO
    ) as log_writer:
        _clear_cache_dict_internal(start_stage, cache_dict, chrom_num_list, log_writer)


def _clear_cache_dict_internal(start_stage, cache_dict, chrom_num_list, log_writer):
    if start_stage == "all":
        _clear_cache_dict_internal(
            "gtg", cache_dict, chrom_num_list, log_writer
        )  # Recursion
        for file_name in ["G0_data.memmap", "ss_per_snp.npz"]:
            if cache_dict[0].file_exists(file_name):
                cache_dict[0].remove(file_name, log_writer=log_writer)
    elif start_stage == "gtg":
        _clear_cache_dict_internal(
            "svd", cache_dict, chrom_num_list, log_writer
        )  # Recursion
        for file_name in ["gtg.npz"]:
            log_writer(file_name)
            if cache_dict[0].file_exists(file_name):
                cache_dict[0].remove(file_name, log_writer=log_writer)
    elif start_stage == "svd":
        _clear_cache_dict_internal(
            "postsvd", cache_dict, chrom_num_list, log_writer
        )  # Recursion
        for chrom in chrom_num_list:
            log_writer("svd chrom {0}".format(chrom))
            cache_dict[0].remove_from_cloud_storage("SVinv_etc{0}.npz".format(chrom))
    elif start_stage == "postsvd":
        # Doing the postsvd deletes before the testsnps deletes is faster because it removes whole containers in a single operation.
        for chrom in chrom_num_list:
            log_writer("postsvd: chrom {0}".format(chrom))
            cache_dict[chrom].remove_from_cloud_storage(log_writer=log_writer)
        _clear_cache_dict_internal(
            "testsnps", cache_dict, chrom_num_list, log_writer
        )  # Recursion
    elif start_stage == "testsnps":
        for file_name in ["G0_data.memmap", "ss_per_snp.npz"]:
            log_writer("cloud storage only '{0}'".format(file_name))
            if cache_dict[0].file_exists(file_name):
                cache_dict[0].cloud_storage_only(file_name, log_writer=log_writer)
        for chrom in [0] + chrom_num_list:
            log_writer("cloud storage only '{0}'".format(chrom))
            cache_dict[chrom].cloud_storage_only(log_writer=log_writer)
    else:
        raise Exception("Don't know start_stage='{0}'".format(start_stage))


def _cache_dict_fixup(cache_dict, chrom_list):
    # If a dictionary, then fix up the values. Else, fix up the value and create a dictionary.
    if isinstance(cache_dict, Mapping):
        return {
            k: FileCache._fixup(v, default_subfolder="single_snp_scale")
            for k, v in cache_dict.items()
        }
    else:
        cache_value = FileCache._fixup(cache_dict, default_subfolder="single_snp_scale")
        return {chrom: cache_value for chrom in [0] + chrom_list}


#!!!if is useful, move near pysnptools's mapreduce.py (with better name)
def map_reduceX(input_seq, mapper=_identity, reducer=list, runner=None, name=None):
    """
    Function for running a function on sequence of inputs and running a second function on the results. Can be nested and clusterized.
    For each top-level input, a separate job will be created.


    :param input_seq: a sequence of inputs. The sequence must support the len function and be indexable. e.g. a list, xrange(100)
    :type input_seq: a sequence

    :param mapper: A function to apply to each set of inputs (optional). Defaults to the identity function. (Also see 'mapper')
    :type mapper: a function

    :param reducer: A function to turn the results from the mapper to a single value (optional). Defaults to creating a list of the results.
    :type reducer: a function that takes a sequence

    :param name: A name to be displayed if this work is done on a cluster.
    :type name: a string

    :param runner: a runner, optional: Tells how to run locally, multi-processor, or on a cluster.
        If not given, the function is run locally.
    :type runner: a runner.

    :rtype: The results from the reducer.

    """
    if runner is None:
        runner = Local()

    if name == None:
        name = str(distributable_list[0]) or ""
        if len(distributable_list) > 1:
            name += "-etc"

    with _MapReduce._dyn_vars(is_in_nested=True):
        distributable_list = [mapper(input) for input in input_seq]
    if hasattr(runner, "run_list"):
        return runner.run_list(distributable_list, reducer=reducer, name=name)
    else:
        result_list = [
            runner.run(distributable) for distributable in distributable_list
        ]
        top_level_result = reducer(result_list)
        return top_level_result


if __name__ == "__main__":
    import doctest
    import logging

    # logging.getLogger().setLevel(logging.WARN)

    doctest.ELLIPSIS_MARKER = "-etc-"
    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    doctest.ELLIPSIS_MARKER = "..."
    print("done")



import numpy as np
import scipy.optimize as opt


def minimize1D(
    f,
    evalgrid=None,
    nGrid=10,
    minval=0.0,
    maxval=0.99999,
    verbose=False,
    brent=True,
    check_boundaries=True,
    resultgrid=None,
    return_grid=False,
):
    """
    minimize a function f(x) in the grid between minval and maxval.
    The function will be evaluated on a grid and then all triplets,
    where the inner value is smaller than the two outer values are optimized by
    Brent's algorithm.
    --------------------------------------------------------------------------
    Input:
    f(x)    : callable target function
    evalgrid: 1-D array prespecified grid of x-values
    nGrid   : number of x-grid points to evaluate f(x)
    minval  : minimum x-value for optimization of f(x)
    maxval  : maximum x-value for optimization of f(x)
    brent   : boolean indicator whether to do Brent search or not.
              (default: True)
    --------------------------------------------------------------------------
    Output list:
    [xopt, f(xopt)]
    xopt    : x-value at the optimum
    f(xopt) : function value at the optimum
    --------------------------------------------------------------------------
    """
    # evaluate the target function on a grid:
    if verbose:
        print("evaluating target function on a grid")
    if evalgrid is not None and brent:  # if brent we need to sort the input values
        i_sort = evalgrid.argsort()
        evalgrid = evalgrid[i_sort]
    if resultgrid is None:
        [evalgrid, resultgrid] = evalgrid1D(
            f, evalgrid=evalgrid, nGrid=nGrid, minval=minval, maxval=maxval
        )

    i_currentmin = resultgrid.argmin()
    minglobal = (evalgrid[i_currentmin], resultgrid[i_currentmin])
    if brent:  # do Brent search in addition to rest?
        if check_boundaries:
            if verbose:
                print(
                    "checking grid point boundaries to see if further search is required"
                )
            if (
                resultgrid[0] < resultgrid[1]
            ):  # if the outer boundary point is a local optimum expand search bounded between the grid points
                if verbose:
                    print(
                        "resultgrid[0]<resultgrid[1]--> outer boundary point is a local optimum expand search bounded between the grid points"
                    )
                minlocal = opt.fminbound(f, evalgrid[0], evalgrid[1], full_output=True)
                if minlocal[1] < minglobal[1]:
                    if verbose:
                        print("found a new minimum during grid search")
                    minglobal = minlocal[0:2]
            if (
                resultgrid[-1] < resultgrid[-2]
            ):  # if the outer boundary point is a local optimum expand search bounded between the grid points
                if verbose:
                    print(
                        "resultgrid[-1]<resultgrid[-2]-->outer boundary point is a local optimum expand search bounded between the grid points"
                    )
                minlocal = opt.fminbound(
                    f, evalgrid[-2], evalgrid[-1], full_output=True
                )
                if minlocal[1] < minglobal[1]:
                    if verbose:
                        print("found a new minimum during grid search")
                    minglobal = minlocal[0:2]
        if verbose:
            print("exploring triplets with brent search")
        onebrent = False
        for i in range(
            resultgrid.shape[0] - 2
        ):  # if any triplet is found, where the inner point is a local optimum expand search
            if (resultgrid[i + 1] < resultgrid[i + 2]) and (
                resultgrid[i + 1] < resultgrid[i]
            ):
                onebrent = True
                if verbose:
                    print("found triplet to explore")
                minlocal = opt.brent(
                    f,
                    brack=(evalgrid[i], evalgrid[i + 1], evalgrid[i + 2]),
                    full_output=True,
                )
                if minlocal[1] < minglobal[1]:
                    minglobal = minlocal[0:2]
                    if verbose:
                        print("found new minimum from brent search")
    if return_grid:
        return (minglobal[0], minglobal[1], evalgrid, resultgrid)
    else:
        return minglobal


def evalgrid1D(f, evalgrid=None, nGrid=10, minval=0.0, maxval=0.99999, dimF=0):
    """
    evaluate a function f(x) on all values of a grid.
    --------------------------------------------------------------------------
    Input:
    f(x)    : callable target function
    evalgrid: 1-D array pre-specified grid of x-values
    nGrid   : number of x-grid points to evaluate f(x)
    minval  : minimum x-value for optimization of f(x)
    maxval  : maximum x-value for optimization of f(x)
    --------------------------------------------------------------------------
    Output:
    evalgrid    : x-values
    resultgrid  : f(x)-values
    --------------------------------------------------------------------------
    """
    if evalgrid is None:
        step = (maxval - minval) / (nGrid)
        evalgrid = np.arange(minval, maxval + step, step)
    if dimF:
        resultgrid = np.ones((evalgrid.shape[0], dimF)) * 9999999999999.0
    else:
        resultgrid = np.ones(evalgrid.shape[0]) * 9999999999999.0
    for i in range(evalgrid.shape[0]):
        fevalgrid = f(evalgrid[i])

        is_real = False
        try:
            is_real = np.isreal(fevalgrid).all()
        except Exception:
            is_real = np.isreal(fevalgrid)
        assert is_real, "function returned imaginary value"

        resultgrid[i] = fevalgrid
    return (evalgrid, resultgrid)



import scipy.stats as ST
import numpy as NP

logit_sigmoid = lambda x: 1.0 / (1.0 + NP.exp(-x))
probit_sigmoid = lambda x: ST.norm.cdf(x)


class Likelihood:
    def __init__(self):
        pass

    # likelihood
    def plain(self, f, y):
        raise NotImplementedError

    # log likelihood
    def log(self, f, y):
        raise NotImplementedError

    # gradient of the log likelihood
    def gradient_log(self, f, y):
        raise NotImplementedError

    # hessian of the log likelihood
    def hessian_log(self, f):
        raise NotImplementedError

    def third_derivative_log(self, f):
        raise NotImplementedError


class ProbitLikelihood(Likelihood):
    def __init__(self):
        Likelihood.__init__(self)

    # likelihood
    def plain(self, f, y):
        return ST.norm._cdf(y * f).prod()

    # log likelihood
    def log(self, f, y):
        return ST.norm._logcdf(y * f).sum()


# Implements p(y|F) = prod_{i=1}^n p(y_i|f_i)
#                   = prod_{i=1}^n logistic(y_i*f_i).
class LogitLikelihood(Likelihood):
    def __init__(self):
        Likelihood.__init__(self)
        # self.sigmoid = lambda x: 1.0 / (1.0 + NP.exp(-x))
        self.sigmoid = logit_sigmoid

    # likelihood
    def plain(self, f, y):
        return self.sigmoid(y * f).prod()

    # log likelihood
    def log(self, f, y):
        yf = y * f
        r = self.sigmoid(yf)
        ok = r > 0.0
        r[ok] = NP.log(r[ok])
        r[~ok] = yf[~ok]
        return r.sum()

    # gradient of the log likelihood
    def gradient_log(self, f, y):
        return y * self.sigmoid(-y * f)

    # hessian of the log likelihood
    def hessian_log(self, f):
        r = self.sigmoid(f) * self.sigmoid(-f)
        ok = r >= 1e-16
        r[~ok] = 1e-16
        return -r
        # maybe faster, but maybe less precise version
        # pi = self.sigmoid(f)
        # return -pi*(1.0-pi)

    def third_derivative_log(self, f):
        pi = self.sigmoid(f)
        return -pi * self.sigmoid(-f) * (1.0 - 2.0 * pi)
        # maybe faster, but maybe less precise version
        # return -pi*(1.0-pi)*(1.0-2.0*pi)

    def intOverGauss(self, mu, sig2):
        # horta201306 we approximate \int logistic(x) N(x, mu, sig2) dx
        # by observing that logistic(x) is cdf of the standard logistic distribution
        # and that this distribution is similar in shape with the normal one
        # so we approximate the logistic pdf by a mixture of five normal
        # distributions, and then calculate the integral
        # we adopted the coefficients c and lambd from Rasmussen and Williams' GP toolbox
        # for Matlab.

        c = NP.array(
            [
                1.146480988574439e02,
                -1.508871030070582e03,
                2.676085036831241e03,
                -1.356294962039222e03,
                7.543285642111850e01,
            ]
        )
        lambd = NP.array([0.44, 0.41, 0.40, 0.39, 0.36])

        sigs = 1.0 / (lambd * NP.sqrt(2))

        z = mu / (sigs * NP.sqrt(1.0 + sig2 / sigs**2))

        return NP.sum(c * ST.norm.cdf(z))



exactUpdate = False
logdelta = np.log(16)
topKbyLinReg = 16

if topKbyLinReg != None:
    outFile = "examples/ScanOSP.exact%d.nSnps%d.logdelta%.2f.out.txt" % (
        exactUpdate,
        topKbyLinReg,
        logdelta,
    )
else:
    outFile = "examples/ScanOSP.exact%d.all.logdelta%.2f.out.txt" % (
        exactUpdate,
        logdelta,
    )

obj = ScanOSP(
    phenoFile="examples/toydata.phe",
    bedFile="examples/toydata",
    windowByPosition=100,
    logdelta=logdelta,
    topKbyLinReg=topKbyLinReg,
    exactUpdate=exactUpdate,
    # prefilterNumSnps = 1000,
    # extractSim = 'examples/LmmGWAS.snps.txt',
    nFolds=5,
    nCV=3,
    outFile=outFile,
    doPlots=True,
    doDebugging=True,
)



from fastlmm.feature_selection import PerformSelectionDistributable as psd
import fastlmm.util.preprocess as up
import numpy as np


def set_Gexclude(G_exclude, G1, i_exclude):
    if G_exclude is None and i_exclude is not None and i_exclude.sum() > 0:
        assert self.G0 is not None, "i_exclude without SNPs to exclude"
        G = np.hstack((self.G0[:, i_exclude], G1))
        i_G1 = np.ones(G.shape[1], dtype="bool")
        n_exclude = i_exclude.sum()
        i_G1[0:n_exclude] = False
    elif G_exclude is not None:
        G = np.hstack((G_exclude, G1))
        i_G1 = np.ones(G.shape[1], dtype="bool")
        n_exclude = G_exclude.shape[1]
        i_G1[0:n_exclude] = False
    else:
        G = G1
        i_G1 = np.ones(G.shape[1], dtype="bool")
        n_exclude = 0
    return G, i_G1, n_exclude


def set_snps0(SNPs0, sample_size, i_exclude=None, forcefullrank=False, blocksize=10000):
    """
    In full rank case, loads up the SNPs in blocks, and construct the kernel.
    In low rank case, loads up all SNPs in to memory
    """
    if SNPs0 is None:
        return None, None
    if "K" in SNPs0:
        K0 = SNPs0["K"]
        G0 = None
    elif "data" in SNPs0:
        K0 = None
        G0 = SNPs0["data"]["snps"]
    else:
        # full rank
        if len(SNPs0["snp_set"]) > sample_size or forcefullrank:  # N = Y.shape[0]
            SNPs0["K"] = psd.build_kernel_blocked(
                snpreader=SNPs0["reader"],
                snp_idx=SNPs0["snp_set"].to_index,
                blocksize=blocksize,
                allowlowrank=forcefullrank,
            )
            K0 = SNPs0["K"]
            G0 = None
        else:
            # low rank
            K0 = None
            SNPs0["data"] = SNPs0["snp_set"].read()
            SNPs0["data"]["snps"] = up.standardize(SNPs0["data"]["snps"])
            G0 = SNPs0["data"]["snps"]

    # lrt_up should never do exclusion, because set_snps0 should only get called once, in run_once, without exclusion
    # exclude. So this is only for score test and lrt.
    if i_exclude is not None:
        if K0 is not None:
            # Also note in the full rank case with exclusion, for score, one could in principle use low rank updates to make it faster,
            # when the number of excluded SNPs is small: it wold be cubic in num_excluded * num_inner*num_outer iterations, versus now
            # where it is cubic in N in the outer loop only once
            K_up = psd.build_kernel_blocked(
                snpreader=SNPs0["reader"],
                snp_idx=np.array(SNPs0["snp_set"].to_index)[i_exclude],
                blocksize=blocksize,
                allowlowrank=forcefullrank,
            )
            K0 = K0 - K_up
        elif G0 is not None:
            G0 = G0[:, ~i_exclude]
        num_snps = SNPs0["num_snps"] - i_exclude.sum()
    else:
        num_snps = SNPs0["num_snps"]
    # intersect data?

    # normalize:
    if K0 is not None:
        K0 = K0 / num_snps  # K0.diagonal().mean()
    elif G0 is not None:
        G0 = G0 / np.sqrt(
            num_snps
        )  # (G0*G0).mean() ) # computes the sqrt of the mean of the diagonal of K=GG^T; *  means pointwise multiplication
    return G0, K0



import sys
import os
import gzip
import re
import numpy as np

writesmall = 0

datadir = ""
base = ""
tbi = base + ".tbi"
basef = os.path.join(datadir, base)
basefsmall = "chr22.small.txt.gz"
basefsmall = "chr22.small.txt"
baseftbi = basef + ".tbi"

if writesmall:
    file = gzip.GzipFile(basef, "r")
    # outfsmall = gzip.GzipFile(basefsmall,'w')
    outfsmall = open(basefsmall, "w")
else:
    file = open(basefsmall, "r")
line = file.readline()
if writesmall:
    outfsmall.write(line)
if line != "##fileformat=VCFv4.1\n":
    print("wrong format")
iline = 0
header = ""
while line[0:2] == "##":
    header = header + line
    line = file.readline()
    if writesmall:
        outfsmall.write(line)
    iline += 1
if line[0:6] == "#CHROM":
    fields = line[1::].strip().split("\t")
    indID = fields[9::]
    fields = fields[0:9]


famfile = open("test.fam", "w")
for id in indID:
    outline = "0 " + id + " 0 0 0 -9\n"
    famfile.write(outline)
famfile.close()
ninds = len(indID)

line = file.readline()
if writesmall:
    outfsmall.write(line)

bedfile = open("test.bed", "wb")
bimfile = open("test.bim", "w")

# init bedfile
bedfile.write(b"l\x1b")  # bed header
bedfile.write(b"\x01")
# genomap=[]
# geno = []
iline = 0
if writesmall:
    maxline = 100
else:
    maxline = 10000000
while (iline < maxline) and (line != ""):
    vals = line.strip().split("\t")
    if vals[6] == "PASS":
        linebim = (
            vals[0]
            + " "
            + vals[2]
            + " 0 "
            + vals[1]
            + " "
            + vals[3]
            + " "
            + vals[4]
            + "\n"
        )
        bimfile.write(linebim)
        # genomap.append(vals[0:9])
        # geno.append(vals[9::])

        # write bedfile entries
        nbytes = np.ceil(ninds / 4.0)
        ndone = 0
        for i in range(int(np.ceil(ninds / 4.0))):
            num = min(4, ninds - ndone)
            geno = vals[9 + i * 4 : 9 + (i) * 4 + num]
            byte = 0
            # pdb.set_trace()
            for j in range(num):

                if geno[j][0:3] == "0|0":
                    # byte += 0*(2**(2*j))
                    pass
                elif geno[j][0:3] == "0|1" or geno[j][0:3] == "1|0":
                    byte += 2 * (2 ** (2 * j))
                elif geno[j][0:3] == "1|1":
                    byte += 3 * (2 ** (2 * j))
                else:
                    # pdb.set_trace()
                    byte += 1 * (2 ** (2 * j))
                pass

            bytestr = "%c" % byte
            bedfile.write(bytestr)
            ndone += num
            pass
    line = file.readline()
    if writesmall:
        outfsmall.write(line)
    iline += 1
file.close()
if writesmall:
    outfsmall.close()
bimfile.close()
bedfile.close()

# filetbi = open(baseftbi,'r')
# filetbigz = gzip.GzipFile(baseftbi,'r')



import numpy as NP
import numpy as np
import scipy.linalg as LA
import scipy.optimize as opt
import scipy.stats as ST
import scipy.special as SS
from fastlmm.util.mingrid import *
from fastlmm.util.util import *
import time

from bin2kernel import Bin2Kernel
from bin2kernel import makeBin2KernelAsEstimator
from bin2kernel import Bin2KernelLaplaceLinearN
from bin2kernel import getFastestBin2Kernel
from bin2kernel import Bin2KernelEPLinearN


def rotate(A, rotationMatrix, transposeRot=True):
    """
    Apply an eigenvalue decomposition rotation matrix (Assummes that the rotation-matrix has orthogonal cols)

    takes care of low rank structure in A

    result = eig[1].T * A
    """
    [Nr, k] = rotationMatrix.shape
    [N, M] = A.shape
    if Nr != N:
        raise Exception(
            "rotation and A are not aligned A.shape =  [%i,%i], rotation.shape = [%i,%i]"
            % (N, M, Nr, k)
        )
    if k > Nr:
        raise Exception(
            "rotation matrix has more columns than rows =  [%i,%i], rotation.shape = [%i,%i]"
            % (N, M, Nr, k)
        )

    res = rotationMatrix.T.dot(A)

    if k < Nr:
        resLowRank = A - rotationMatrix.dot(res)
        res = [res, resLowRank]
    return res


def rotSymm(A, eig, delta=1.0, gamma=1.0, exponent=-0.5, forceSymm=True):
    """
    do the rotation with K_0 to get a matrix square root
    """
    [N, M] = A.shape
    [Nr, k] = eig[1].shape
    if Nr != N:
        raise Exception(
            "rotation and A are not aligned A.shape =  [%i,%i], rotation.shape = [%i,%i]"
            % (N, M, Nr, k)
        )
    if symmetric and N != M:
        raise Exception(
            "A is not symmetric. A.shape =  [%i,%i], rotation.shape = [%i,%i]"
            % (N, M, Nr, k)
        )
    if k > Nr:
        raise Exception(
            "rotation matrix has more columns than rows =  [%i,%i], rotation.shape = [%i,%i]"
            % (N, M, Nr, k)
        )

    res = eig[1].T.dot(A)
    deltaPow = myPow(delta, exponent, minVal)
    diag = eig[0] * gamma + delta
    diag = myPow(diag, exponent, minVal=0.0)
    if Nr > k:
        diag -= deltaPow
        res = dotDiag(res, diag)
        res = A - eig[1].dot(res)
    else:
        res = dotDiag(res, diag)
        if forceSymm:
            res = eig[1].dot(res)
    return res


def dotDiag(A, diag, minVal=None, symmetric=False, exponent=1.0):
    """
    Multiply by a diagonal matrix build from

    result = diag.dot(A)
    """
    [N, M] = A.shape
    [Nd] = diag.shape
    if symmetric and N != M:
        raise Exception(
            "A is not symmetric. A.shape =  [%i,%i], rotation.shape = [%i,%i]"
            % (N, M, Nr, k)
        )
    if Nd != N:
        raise Exception(
            "Matrices misaligned A.shape =  [%i,%i], diag.shape = [%i,%i]"
            % (N, M, Nr, k)
        )

    diagcorr = myPow(diag, exponent, minVal)

    result = A * NP.lib.stride_tricks.as_strided(
        diagcorr, (diagcorr.size, A.shape[1]), (diagcorr.itemsize, 0)
    )
    if symmetric:
        result = dotDiag(
            A=A.T,
            diag=diagcorr,
            checkPos=False,
            minVal=minVal,
            symmetric=False,
            exponent=1.0,
        ).T
    return result


def myPow(A, exponent, minVal=None):
    """
    compute a power, with efficient computation of common powers
    """
    if exponent == 1.0:
        Aexp = A
    elif exponent == -1.0:
        Aexp = 1.0 / A
    elif exponent == 2.0:
        Aexp = A * A
    elif exponent == 0.5:
        Aexp = NP.sqrt(A)
    elif exponent == -0.5:
        Aexp = 1.0 / NP.sqrt(A)
    elif exponent == -2.0:
        Aexp = 1.0 / (A * A)
    else:
        Aexp = NP.power(A, exponent)
    if minVal is not None:
        i_Null = A <= minVal
        Aexp[i_Null] = 0.0
    return Aexp


class lmm2k(object):
    """
    linear mixed model with up to two kernels
    N(y | X*beta ; sigma2(gamma0*K0 + gamma1*K1 + I ) ),
    where
    K0 = G0*G0^T
    K1 = G1*G1^T
    """

    __slots__ = [
        "G0",
        "G1",
        "Y",
        "X",
        "K0",
        "K1",
        "K",
        "U",
        "S",
        "UX",
        "Uy",
        "UUX",
        "UW",
        "UUW",
        "UUy",
        "pos0",
        "pos1",
        "gamma0",
        "gamma1",
        "delta",
        "exclude_idx",
        "forcefullrank",
        "numcalls",
        "Xstar",
        "Kstar",
        "Kstar_star",
        "UKstar",
        "UUKstar",
        "Gstar",
    ]

    def __init__(self, forcefullrank=False):
        """
        Input:
        forcefullrank   : if True, then the code always computes K and runs cubically
                            (False)
        """
        self.X = None
        self.Y = None
        self.G0 = None
        self.G1 = None
        self.K0 = None
        self.K1 = None
        self.gamma0 = 1.0
        self.gamma1 = 1.0
        self.delta = 1.0
        self.eig0 = None
        self.eig1 = None
        self.Yrot = None
        self.Xrot = None
        self.K1rot = None
        self.G1rot = None

    def setX(self, X):
        """
        set the fixed effects X (covariates).
        The Kernel has to be set in advance by first calling setG() or setK().
        --------------------------------------------------------------------------
        Input:
        X       : [N*D] 2-dimensional array of covariates
        --------------------------------------------------------------------------
        """
        self.X = X
        self.Xrot = None

    def setY(self, Y):
        """
        set the phenotype y.
        The Kernel has to be set in advance by first calling setG() or setK().
        --------------------------------------------------------------------------
        Input:
        Y       : [NxP] P-dimensional array of phenotype values
        --------------------------------------------------------------------------
        """
        self.Y = Y
        self.Yrot = None

    def setG0(self, G0):
        """
        set the Kernel K0 from G0.
        This has to be done before setting the data setX() and setY().
        ----------------------------------------------------------------------------
        Input:
        G0              : [N*k0] array of random effects
        -----------------------------------------------------------------------------
        """
        k = G0.shape[1]
        N = G0.shape[0]
        self.eig0 = None
        self.eig1 = None
        self.K1rot = None
        self.Yrot = None
        self.Xrot = None

        if (not self.forcefullrank) and (k < N):
            self.G0 = G0
        else:
            K0 = G0.dot(G0.T)
            self.setK0(K0=K0)

    def setK0(self, K0):
        """
        set the background Kernel K0.
        --------------------------------------------------------------------------
        Input:
        K0 : [N*N] array, random effects covariance (positive semi-definite)
        --------------------------------------------------------------------------
        """
        self.K0 = K0
        self.G1 = None
        self.eig0 = None
        self.eig1 = None
        self.K1rot = None
        self.Yrot = None
        self.Xrot = None
        self.G1rot = None
        self.K1rot = None

    def setG1(self, G1):
        """
        set the Kernel K1 from G1.
        ----------------------------------------------------------------------------
        Input:
        G0              : [N*k0] array of random effects
        -----------------------------------------------------------------------------
        """
        k = self.G1.shape[1]
        N = self.G0.shape[0]
        self.eig1 = None
        self.G1rot = None
        if (not self.forcefullrank) and (k < N):
            # it is faster using the eigen decomposition of G.T*G but this is more accurate
            self.G1 = G1
        else:
            K1 = self.G1.dot(self.G1.T)
            self.setK1(K1=K1)
        pass

    def setK1(self, K1):
        """
        set the foreground Kernel K1.
        This has to be done before setting the data setX() and setY().
        --------------------------------------------------------------------------
        Input:
        K1 : [N*N] array, random effects covariance (positive semi-definite)
        --------------------------------------------------------------------------
        """
        self.eig1 = None
        self.G1 = None
        self.K1 = K1
        self.K1rot = None
        self.G1rot = None

    def setVariances(self, gamma0=None, gamma1=None, delta=None):
        if gamma0 is not None:
            # background model changed, foreground rotation (U1,S1) changes:
            self.gamma0 = gamma0
            self.eig1 = None
            self.Xrot = None
            self.Yrot = None
            self.K1rot = None
            self.G1rot = None

        if delta is not None:
            # background model changed, foreground rotation (U1,S1) changes:
            self.delta = delta
            self.eig1 = None
            self.Xrot = None
            self.Yrot = None
            self.K1rot = None
            self.G1rot = None

        if gamma1 is not None:
            # foreground model changed
            self.gamma1 = gamma1

    def getEig1(self):
        """ """
        if self.eig1 is None:
            # compute eig1
            if self.K1 is not None:
                if self.K1rot is None:
                    self.K1rot = rotSymm(
                        self.K1,
                        eig=self.eig0,
                        exponent=-0.5,
                        gamma=self.gamma0,
                        delta=self.delta,
                        forceSymm=False,
                    )
                    self.K1rot = rotSymm(
                        self.K1rot.T,
                        eig=self.eig0,
                        exponent=-0.5,
                        gamma=self.gamma0,
                        delta=self.delta,
                        forceSymm=False,
                    )
                self.eig1 = LA.eigh(self.K1rot)
            elif self.G1 is not None:
                [N, k] = self.G1.shape
                if self.G1rot is None:
                    self.G1rot = rotSymm(
                        self.G1,
                        eig=self.eig0,
                        exponent=-0.5,
                        gamma=self.gamma0,
                        delta=self.delta,
                        forceSymm=False,
                    )

                try:
                    [U, S, V] = LA.svd(self.G1rot, full_matrices=False)
                    self.eig1 = [S * S, U]
                except LA.LinAlgError:  # revert to Eigenvalue decomposition
                    print(
                        "Got SVD exception, trying eigenvalue decomposition of square of G. Note that this is a little bit less accurate"
                    )
                    [S_, V_] = LA.eigh(self.G1rot.T.dot(self.G1rot))
                    S_nonz = S_ > 0.0
                    S1 = S_[S_nonz]
                    U1 = self.G1rot.dot(V_[:, S_nonz] / np.sqrt(S1))
                    self.eig1 = [S1, U1]
        return self.eig1

    def getEig0(self):
        """ """
        if self.eig0 is None:
            # compute eig0
            if self.K0 is not None:
                self.eig1 = LA.eigh(self.K0)
            elif self.G1 is not None:
                [N, k] = self.G0.shape
                try:
                    [U, S, V] = LA.svd(self.G0, full_matrices=False)
                    self.eig0 = [S * S, U]
                except LA.LinAlgError:  # revert to Eigenvalue decomposition
                    print(
                        "Got SVD exception, trying eigenvalue decomposition of square of G. Note that this is a little bit less accurate"
                    )
                    [S_, V_] = LA.eigh(self.G0.T.dot(self.G0))
                    S_nonz = S_ > 0.0
                    S0 = S_[S_nonz]
                    U0 = self.G0.dot(V_[:, S_nonz] / np.sqrt(S0))
                    self.eig0 = [S0, U0]
        return self.eig1

    def set_exclude_idx(self, idx):
        """
        Set the indices of SNPs to be removed
        --------------------------------------------------------------------------
        Input:
        idx  : [k_up: number of SNPs to be removed] holds the indices of SNPs to be removed
        --------------------------------------------------------------------------
        """
        self.exclude_idx = idx

    def findGamma1givenGamma0(
        self, gamma0=0.0, nGridGamma1=10, minGamma1=0.0, maxGamma1=10000.0, **kwargs
    ):
        """
        Find the optimal h2 for a given K. Note that this is the single kernel case. So there is no a2.
        (default maxH2 value is set to a value smaller than 1 to avoid loss of positive definiteness of the final model covariance)
        --------------------------------------------------------------------------
        Input:
        nGridH2 : number of h2-grid points to evaluate the negative log-likelihood at
        minH2   : minimum value for h2 optmization
        maxH2   : maximum value for h2 optmization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal h2
        --------------------------------------------------------------------------
        """
        self.setVariances(gamma0=gamma0)
        resmin = [None]

        def f(x, resmin=resmin, **kwargs):
            # self.setVariances(gamma1=x)
            res = self.nLLeval(gamma1=x, **kwargs)
            if (resmin[0] is None) or (res["nLL"] < resmin[0]["nLL"]):
                resmin[0] = res
                self.setGamma1[x]
            return res["nLL"]

        min = minimize1D(f=f, nGrid=nGridGamma1, minval=minGamma1, maxval=maxGamma1)
        return resmin[0]

    def findGammas(
        self,
        nGridGamma0=10,
        minGamma0=0.0,
        maxGamma0=10000.0,
        nGridGamma1=10,
        minGamma1=0.0,
        maxGamma1=10000.0,
        verbose=False,
        **kwargs
    ):
        """
        Find the optimal gamma0 and gamma1, such that K=gamma0*K0+gamma1*K1. Performs a double loop optimization (could be expensive for large grid-sizes)
        --------------------------------------------------------------------------
        Input:
        nGridGamma0 : number of a2-grid points to evaluate the negative log-likelihood at
        minGamma0   : minimum value for a2 optmization
        maxGamma0   : maximum value for a2 optmization
        nGridGamma1 : number of h2-grid points to evaluate the negative log-likelihood at
        minGamma1   : minimum value for h2 optmization
        maxGamma1   : maximum value for h2 optmization
        --------------------------------------------------------------------------
        Output:
        dictionary containing the model parameters at the optimal gamma0 and gamma1
        --------------------------------------------------------------------------
        """
        self.numcalls = 0
        resmin = [None]

        def f(
            x,
            resmin=resmin,
            nGridGamma1=nGridGamma1,
            minGamma1=minGamma1,
            maxGamma1=maxGamma1,
            **kwargs
        ):
            self.numcalls += 1
            t0 = time.time()
            res = self.findGamma1givenGamma0(
                gamma0=x,
                nGridGamma1=nGridGamma1,
                minGamma1=minGamma1,
                maxGamma1=maxGamma1,
                **kwargs
            )
            if (resmin[0] is None) or (res["nLL"] < resmin[0]["nLL"]):
                resmin[0] = res
            t1 = time.time()
            # print "one objective function call took %.2f seconds elapsed" % (t1-t0)
            # import pdb; pdb.set_trace()
            return res["nLL"]

        if verbose:
            print("findGammas")
        min = minimize1D(
            f=f, nGrid=nGridGamma1, minval=minGamma1, maxval=maxGamma1, verbose=False
        )
        # print "numcalls to innerLoopTwoKernel= " + str(self.numcalls)
        return resmin[0]

    def nLLeval(self, REML=True, gamma1=None, dof=None, scale=1.0):
        """
        evaluate -ln( N( U^T*y | U^T*X*beta , h2*S + (1-h2)*I ) ),
        where ((1-a2)*K0 + a2*K1) = USU^T
        --------------------------------------------------------------------------
        Input:
        h2      : mixture weight between K and Identity (environmental noise)
        REML    : boolean
                  if True   : compute REML
                  if False  : compute ML
        dof     : Degrees of freedom of the Multivariate student-t
                        (default None uses multivariate Normal likelihood)
        logdelta: log(delta) allows to optionally parameterize in delta space
        delta   : delta     allows tomoptionally parameterize in delta space
        scale   : Scale parameter the multiplies the Covariance matrix (default 1.0)
        --------------------------------------------------------------------------
        Output dictionary:
        'nLL'       : negative log-likelihood
        'sigma2'    : the model variance sigma^2
        'beta'      : [D*1] array of fixed effects weights beta
        'h2'        : mixture weight between Covariance and noise
        'REML'      : True: REML was computed, False: ML was computed
        'a2'        : mixture weight between K0 and K1
        'dof'       : Degrees of freedom of the Multivariate student-t
                        (default None uses multivariate Normal likelihood)
        'scale'     : Scale parameter that multiplies the Covariance matrix (default 1.0)
        --------------------------------------------------------------------------
        """
        if gamma1 is None:
            gamma1 = self.gamma1
        if gamma1 < 0.0:
            return {
                "nLL": 3e20,
                "gamma1": gamma1,
                "delta": self.delta,
                "gamma0": self.gamma0,
                "dof": dof,
                "REML": REML,
                "scale": scale,
            }
        k = self.S1.shape[0]
        N = self.Y.shape[0]
        D = self.X.shape[1]

        Sd = (gamma1 * self.S1 + 1.0) * scale
        UXS = dotDiag(self.UX, Sd, minVal=0.0, symmetric=False, exponent=-1.0)
        UYS = dotDiag(self.UY, Sd, minVal=0.0, symmetric=False, exponent=-1.0)

        XKX = UXS.T.dot(self.UX)
        XKy = UXS.T.dot(self.Uy)
        yKy = UyS.T.dot(self.Uy)

        logdetK = np.log(Sd).sum()

        if k < N:  # low rank part
            # determine normalization factor
            denom = 1.0 * scale

            XKX += self.UUX.T.dot(self.UUX) / (denom)
            XKy += self.UUX.T.dot(self.UUy) / (denom)
            yKy += self.UUy.T.dot(self.UUy) / (denom)
            logdetK += (N - k) * np.log(denom)

        # proximal contamination (see Supplement Note 2: An Efficient Algorithm for Avoiding Proximal Contamination)
        # available at: http://www.nature.com/nmeth/journal/v9/n6/extref/nmeth.2037-S1.pdf
        # exclude SNPs from the RRM in the likelihood evaluation

        if len(self.exclude_idx) > 0:
            raise Exception("not implemented")
            num_exclude = len(self.exclude_idx)

            # consider only excluded SNPs
            G_exclude = self.G[:, self.exclude_idx]

            self.UW = self.U.T.dot(G_exclude)  # needed for proximal contamination
            UWS = self.UW / NP.lib.stride_tricks.as_strided(
                Sd, (Sd.size, num_exclude), (Sd.itemsize, 0)
            )
            assert UWS.shape == (k, num_exclude)

            WW = NP.eye(num_exclude) - UWS.T.dot(self.UW)
            WX = UWS.T.dot(self.UX)
            Wy = UWS.T.dot(self.Uy)
            assert WW.shape == (num_exclude, num_exclude)
            assert WX.shape == (num_exclude, D)
            assert Wy.shape == (num_exclude,)

            if k < N:  # low rank part
                self.UUW = G_exclude - self.U.dot(self.UW)

                WW += self.UUW.T.dot(self.UUW) / denom
                WX += self.UUW.T.dot(self.UUX) / denom
                Wy += self.UUW.T.dot(self.UUy) / denom

            # TODO: do cholesky, if fails do eigh
            # compute inverse efficiently
            [S_WW, U_WW] = LA.eigh(WW)

            UWX = U_WW.T.dot(WX)
            UWy = U_WW.T.dot(Wy)
            assert UWX.shape == (num_exclude, D)
            assert UWy.shape == (num_exclude,)

            # compute S_WW^{-1} * UWX
            WX = UWX / NP.lib.stride_tricks.as_strided(
                S_WW, (S_WW.size, UWX.shape[1]), (S_WW.itemsize, 0)
            )
            # compute S_WW^{-1} * UWy
            Wy = UWy / S_WW
            # determinant update
            logdetK += np.log(S_WW).sum()
            assert WX.shape == (num_exclude, D)
            assert Wy.shape == (num_exclude,)

            # perform updates (instantiations for a and b in Equation (1.5) of Supplement)
            yKy += UWy.T.dot(Wy)
            XKy += UWX.T.dot(Wy)
            XKX += UWX.T.dot(WX)

        #######

        [SxKx, UxKx] = LA.eigh(XKX)
        i_pos = SxKx > 1e-10
        beta = np.dot(UxKx[:, i_pos], (np.dot(UxKx[:, i_pos].T, XKy) / SxKx[i_pos]))

        r2 = yKy - XKy.dot(beta)

        if dof is None:  # Use the Multivariate Gaussian
            if REML:
                XX = self.X.T.dot(self.X)
                [Sxx, Uxx] = LA.eigh(XX)
                logdetXX = np.log(Sxx).sum()
                logdetXKX = np.log(SxKx).sum()
                sigma2 = r2 / (N - D)
                nLL = 0.5 * (
                    logdetK
                    + logdetXKX
                    - logdetXX
                    + (N - D) * (np.log(2.0 * np.pi * sigma2) + 1)
                )
            else:
                sigma2 = r2 / (N)
                nLL = 0.5 * (logdetK + N * (np.log(2.0 * np.pi * sigma2) + 1))
            result = {
                "nLL": nLL,
                "sigma2": sigma2,
                "beta": beta,
                "gamma1": gamma1,
                "REML": REML,
                "gamma0": self.gamma0,
                "delta": self.delta,
                "scale": scale,
            }
        else:  # Use multivariate student-t
            if REML:
                XX = self.X.T.dot(self.X)
                [Sxx, Uxx] = LA.eigh(XX)
                logdetXX = np.log(Sxx).sum()
                logdetXKX = np.log(SxKx).sum()

                nLL = 0.5 * (
                    logdetK
                    + logdetXKX
                    - logdetXX
                    + (dof + (N - D)) * np.log(1.0 + r2 / dof)
                )
                nLL += (
                    0.5 * (N - D) * np.log(dof * np.pi)
                    + SS.gammaln(0.5 * dof)
                    - SS.gammaln(0.5 * (dof + (N - D)))
                )
            else:
                nLL = 0.5 * (logdetK + (dof + N) * np.log(1.0 + r2 / dof))
                nLL += (
                    0.5 * N * np.log(dof * np.pi)
                    + SS.gammaln(0.5 * dof)
                    - SS.gammaln(0.5 * (dof + N))
                )
            result = {
                "nLL": nLL,
                "dof": dof,
                "sigma2": sigma2,
                "beta": beta,
                "gamma1": gamma1,
                "REML": REML,
                "gamma0": self.gamma0,
                "delta": self.delta,
                "scale": scale,
            }

        return result



"""
example of how to use feature selection from python (see also command line interface)
"""

import numpy as np
from fastlmm.feature_selection import FeatureSelectionStrategy
from pysnptools.snpreader import Bed

import logging


def runselect(
    bed_fn=None,
    pheno_fn=None,
    strategy=None,
    select_by_ll=True,
    output_prefix=None,
    num_pcs=0,
    random_state=3,
    num_snps_in_memory=1000,
    cov_fn=None,
    k_values=None,
    delta_values=None,
    num_folds=10,
    penalty=0.0,
):
    logging.basicConfig(level=logging.INFO)

    # set up data
    ##############################
    if bed_fn is None:
        bed_fn = Bed("examples/toydata")
        pheno_fn = "examples/toydata.phe"

    # set up grid
    ##############################
    num_steps_delta = 10
    num_steps_k = 5

    # log_2 space and all SNPs
    # k_values = np.logspace(0, 9, base=2, num=num_steps_k, endpoint=True).tolist() + [10000]
    if k_values is None:
        k_values = [
            0,
            1,
            5,
            10,
            20,
            50,
            100,
            500,
            1000,
            2000,
            5000,
            10000,
            456345643256,
        ]  # 100, 110, 120, 130, 140, 150, 160, 170, 180, 190, 200, 210, 220, 230, 240, 250, 260, 270, 280, 290, 300, 400, 500, 1000]
    if delta_values is None:
        delta_values = np.logspace(
            -10, 10, endpoint=True, num=num_steps_delta, base=np.exp(1)
        )
    # delta_values = [np.exp(1), np.exp(2), np.exp(3), np.exp(4), np.exp(5), np.exp(6)]

    if strategy is None:
        strategy = "lmm_full_cv"
        select_by_ll = True
    if 0:
        strategy = "insample_cv"
        select_by_ll = True
    # where to save output
    ##############################
    if output_prefix is None:
        output_prefix = "example_pc%i" % (num_pcs)

    # go!
    fss = FeatureSelectionStrategy(
        bed_fn,
        pheno_fn,
        num_folds,
        random_state=random_state,
        num_pcs=num_pcs,
        num_snps_in_memory=num_snps_in_memory,
        interpolate_delta=False,
        cov_fn=cov_fn,
    )

    best_k, best_delta, best_obj, best_snps = fss.perform_selection(
        k_values,
        delta_values,
        output_prefix=output_prefix,
        select_by_ll=select_by_ll,
        strategy=strategy,
        penalty=penalty,
    )
    res = {
        "best_k": best_k,
        "best_delta": best_delta,
        "best_obj": best_obj,
        "best_snps": best_snps,
    }
    return res


if __name__ == "__main__":
    result = runselect()



import numpy as np
import logging

from fastlmm.feature_selection import FeatureSelectionStrategy, load_snp_data
from pysnptools.snpreader import Bed
import pysnptools.util.pheno as pstpheno
import fastlmm.inference.linear_regression as lin_reg
from pysnptools.snpreader import Hdf5
from pysnptools.snpreader import Dat
from pysnptools.snpreader import Ped

#  sklearn
from sklearn.model_selection import KFold
from pysnptools.standardizer import Unit

from fastlmm.inference import getLMM
import unittest
import os.path
from unittest.mock import patch

import pysnptools.util
from fastlmm.feature_selection.feature_selection_two_kernel import (
    FeatureSelectionInSample,
)
import fastlmm.util.standardizer as stdizer


class TestTwoKernelFeatureSelection(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        self.snp_fn = currentFolder + "/../../tests/datasets/mouse/alldata"
        self.pheno_fn = (
            currentFolder + "/../../tests/datasets/mouse/pheno_10_causals.txt"
        )
        # self.cov_fn = currentFolder + "/examples/toydata.cov"

        # load data
        ###################################################################
        snp_reader = Bed(self.snp_fn, count_A1=False)
        pheno = pstpheno.loadOnePhen(self.pheno_fn)
        # cov = pstpheno.loadPhen(self.cov_fn)

        # intersect sample ids
        snp_reader, pheno = pysnptools.util.intersect_apply([snp_reader, pheno])

        self.G = snp_reader.read(order="C").val
        self.G = stdizer.Unit().standardize(self.G)
        self.G.flags.writeable = False
        self.y = pheno["vals"][:, 0]
        self.y.flags.writeable = False

        # load pcs
        # self.G_cov = cov['vals']
        self.G_cov = np.ones((len(self.y), 1))
        self.G_cov.flags.writeable = False

    def test_regression_lmm(self):

        # invoke fs
        select = FeatureSelectionInSample(
            n_folds=2, max_log_k=6, order_by_lmm=True, measure="mse", random_state=42
        )
        best_k, feat_idx, best_mix, best_delta = select.run_select(
            self.G, self.G, self.y, cov=self.G_cov
        )

        # print results
        print("best_k:", best_k)
        print("best_mix:", best_mix)
        print("best_delta:", best_delta)

        self.assertEqual(best_k, 64)
        self.assertAlmostEqual(best_mix, 0.8621642030968627, places=6)
        self.assertAlmostEqual(best_delta, 0.7255878551207211, places=6)

    def test_regression_lr(self):

        # invoke fs
        select = FeatureSelectionInSample(
            n_folds=2, max_log_k=6, order_by_lmm=False, measure="mse", random_state=42
        )
        best_k, feat_idx, best_mix, best_delta = select.run_select(
            self.G, self.G, self.y, cov=self.G_cov
        )

        # print results
        print("best_k:", best_k)
        print("best_mix:", best_mix)
        print("best_delta:", best_delta)

        self.assertEqual(best_k, 32)
        self.assertAlmostEqual(best_mix, 0.6786566031577429, places=6)
        self.assertAlmostEqual(best_delta, 0.70148446599200931, places=6)


class TestFeatureSelection(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        self.snpreader_bed = Bed(
            currentFolder + "/examples/toydata.5chrom.bed", count_A1=False
        )
        # Creating Hdf5 data ...
        # snpData = self.snpreader_bed.read()
        # Hdf5.write(snpData, currentFolder + "/examples/toydata.snpmajor.hdf5")
        # Hdf5.write(snpData, currentFolder + "/examples/toydata.iidmajor.hdf5",snp_major=False)

        # Creating Dat data ...
        # snpData = self.snpreader_bed.read()
        # Dat.write(snpData, currentFolder + "/examples/toydata.dat")

        ###Creating Ped data ...
        # snpData = self.snpreader_bed.read()
        # Ped.write(snpData, currentFolder + "/examples/toydata.ped")
        # fromPed = Ped(currentFolder + "/examples/toydata").read()
        # self.assertTrue(np.allclose(snpData.val, fromPed.val, rtol=1e-05, atol=1e-05))

        self.snpreader_hdf5 = Hdf5(currentFolder + "/examples/toydata.snpmajor.hdf5")
        self.snpreader_dat = Dat(currentFolder + "/examples/toydata.dat")
        self.snpreader_ped = Ped(currentFolder + "/examples/toydata")
        self.pheno_fn = currentFolder + "/examples/toydata.phe"
        self.pheno_shuffleplus_fn = currentFolder + "/examples/toydata.shufflePlus.phe"

    """
    make sure the used pca yields the same result as standard pca
    def test_pca(self):

        from sklearn.decomposition import PCA, KernelPCA

        print "testing PCA"

        num_steps_delta = 5
        num_folds = 2
        num_pcs = 2
        random_state = 42
        output_prefix = None
        fss = FeatureSelectionStrategy(self.snpreader, self.pheno_fn, num_folds, random_state=random_state, num_pcs=num_pcs, interpolate_delta=True, num_snps_in_memory=20000,count_A1=False)
        fss.run_once()

        pca = PCA(n_components=fss.num_pcs)
        pcs = pca.fit_transform(fss.G)

        for i in xrange(num_pcs):
            
            pc_1 = fss.pcs[:,i]
            pc_2 = pcs[:,i]
            # sign -1 if signs different, 1 else
            sign = np.sign(pc_1[0]) * np.sign(pc_2[0])

            np.testing.assert_array_almost_equal(pc_1, sign*pc_2)
    """

    def test_regression_bed(self):
        self.regression(self.snpreader_bed, self.regular_regression_answers)

    def test_regression_hdf5(self):
        self.regression(self.snpreader_hdf5, self.regular_regression_answers)

    def test_regression_dat(self):
        self.regression(self.snpreader_dat, self.regular_regression_answers)

    def test_regression_ped(self):
        self.regression(self.snpreader_ped, self.regular_regression_answers)

    regular_regression_answers = (22, 20.085536923, 61.2448170241, 0.67586545761317196)
    cov_pca_regression_answers = (22, 20.085536923, 61.8146293815, 0.692761716513)
    cov_pca_insample_cv_regression_answers = (
        22,
        1.6513737331988527,
        63.6062289765,
        0.71724685708485092,
    )

    def test_regression_cov_pcs(self):
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        cov_fn = currentFolder + "/examples/toydata.cov"
        self.regression(
            self.snpreader_bed,
            self.cov_pca_regression_answers,
            cov_fn=cov_fn,
            num_pcs=3,
        )

    def test_regression_cov_pcs_insample_cv(self):
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        cov_fn = currentFolder + "/examples/toydata.cov"
        self.regression(
            self.snpreader_bed,
            self.cov_pca_insample_cv_regression_answers,
            cov_fn=cov_fn,
            num_pcs=3,
            strategy="insample_cv",
            delta=5,
        )

    def regression(
        self,
        snpreader,
        answers,
        cov_fn=None,
        num_pcs=0,
        strategy="lmm_full_cv",
        delta=7,
    ):
        """
        compare against previous results of this code base
        """

        # set up grid
        ##############################
        num_steps_delta = 5
        num_steps_k = 5
        num_folds = 2

        # log_2 space and all SNPs
        k_values = np.array(
            np.logspace(0, 9, base=2, num=num_steps_k, endpoint=True), dtype=np.int64
        ).tolist() + [10000]
        # k_values = np.logspace(0, 9, base=2, num=num_steps_k, endpoint=True).tolist() + [10000]
        delta_values = np.logspace(
            -3, 3, endpoint=True, num=num_steps_delta, base=np.exp(1)
        )
        random_state = 42

        output_prefix = None

        # select by LL
        fss = FeatureSelectionStrategy(
            snpreader,
            self.pheno_fn,
            num_folds,
            random_state=random_state,
            cov_fn=cov_fn,
            num_pcs=num_pcs,
            interpolate_delta=True,
            count_A1=False,
        )
        best_k, best_delta, best_obj, best_snps = fss.perform_selection(
            k_values,
            delta_values,
            strategy,
            output_prefix=output_prefix,
            select_by_ll=True,
            create_pdf=False,
        )

        self.assertEqual(best_k, answers[0])
        self.assertAlmostEqual(best_delta, answers[1], delta)
        self.assertTrue(
            abs(best_obj - answers[2]) < 0.005
        )  # accept a range answers for when standardization is done with doubles, floats, etc

        # select by MSE
        fss = FeatureSelectionStrategy(
            snpreader,
            self.pheno_fn,
            num_folds,
            random_state=random_state,
            cov_fn=cov_fn,
            num_pcs=num_pcs,
            interpolate_delta=True,
            count_A1=False,
        )
        best_k, best_delta, best_obj, best_snps = fss.perform_selection(
            k_values,
            delta_values,
            strategy,
            output_prefix=output_prefix,
            select_by_ll=False,
            create_pdf=False,
        )

        self.assertEqual(best_k, answers[0])
        self.assertAlmostEqual(best_delta, answers[1], delta)
        self.assertAlmostEqual(best_obj, answers[3])

    def test_blocking_bed(self):
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        self.blocking(
            currentFolder + "/examples/toydata.5chrom.bed"
        )  # use string instead of reader, to test that strings work

    def test_blocking_hdf5(self):
        self.blocking(self.snpreader_hdf5)

    def test_blocking_dat(self):
        self.blocking(self.snpreader_dat)

    def test_blocking_ped(self):
        self.blocking(self.snpreader_ped)

    tolerance = 1e-4

    def test_blocking_cov_pcs(self):
        self.blocking_cov_pcs(strategy="lmm_full_cv")

    def getworkingtest_blocking_cov_pcs_insample_cv(self):  #!!!
        self.blocking_cov_pcs(strategy="insample_cv")

    @staticmethod
    def reference_file(outfile):
        #!!similar code elsewhere
        import platform

        os_string = platform.platform()

        file_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "..", "..", "tests"
        )
        windows_fn = file_path + "/expected-Windows/" + outfile
        assert os.path.exists(windows_fn), "Can't find file '{0}'".format(windows_fn)
        debian_fn = file_path + "/expected-debian/" + outfile
        if not os.path.exists(
            debian_fn
        ):  # If reference file is not in debian folder, look in windows folder
            debian_fn = windows_fn

        if "debian" in os_string or "Linux" in os_string:
            if "Linux" in os_string:
                logging.warning(
                    "comparing to Debian output even though found: %s" % os_string
                )
            return debian_fn
        else:
            if "Windows" not in os_string:
                logging.warning(
                    "comparing to Windows output even though found: %s" % os_string
                )
            return windows_fn

    def blocking_cov_pcs(self, strategy):
        currentFolder = os.path.dirname(os.path.realpath(__file__))
        cov_fn = currentFolder + "/examples/toydata.cov"
        output_dir = "tmp"
        try:
            os.mkdir(output_dir)
        except:
            pass
        output_dir = output_dir + "/feature_selection"
        try:
            os.mkdir(output_dir)
        except:
            pass
        self.blocking(
            self.snpreader_bed,
            cov_fn,
            num_pcs=3,
            strategy=strategy,
            output_prefix=os.path.join(output_dir, strategy),
        )

        for outfile in os.listdir(output_dir):
            referenceOutfile = TestFeatureSelection.reference_file(outfile)

            import fastlmm.util.util as ut

            if (
                outfile.lower().endswith(".pdf")
                or outfile == "output_prefix_report.txt"
                or outfile.lower().endswith("_k_pcs.txt")
            ):
                self.assertTrue(os.path.exists(referenceOutfile))
            else:
                delimiter = "," if outfile.lower().endswith(".csv") else "\t"
                tmpOutfile = os.path.join(output_dir, outfile)
                out, msg = ut.compare_mixed_files(
                    tmpOutfile, referenceOutfile, self.tolerance, delimiter=delimiter
                )
                if not out:
                    from pathlib import Path

                    msg = f"{msg}: tempOutfile={Path(tmpOutfile).read_text()} referenceOutfile={Path(referenceOutfile).read_text()}"
                self.assertTrue(
                    out,
                    "msg='{0}', ref='{1}', tmp='{2}'".format(
                        msg, referenceOutfile, tmpOutfile
                    ),
                )

    def blocking(
        self,
        snpreader,
        cov_fn=None,
        num_pcs=0,
        output_prefix=None,
        strategy="lmm_full_cv",
    ):
        """
        compare three different cases:

        To control memory use, we've introduced a parameter called "num_snps_in_memory", which defaults to 100000.
        Here are the interesting cases to consider (and choose num_snps_in_memory accordingly):

        1) num_snps_in_memory > total_num_snps

           In this case, the same code as before should be
           executed (except the kernel matrix on all SNPs is now cached).


        2) num_snps_in_memory < total_num_snps
            num_snps_in_memory > k (excluding all_snps)

            Here, the linear regression will be blocked,
            while the data for cross-validation is cached,
            saving time for loading and re-indexing.


        3) num_snps_in_memory < total_num_snps
            num_snps_in_memory < k (excluding all_snps)

            Finally, both operations - linear regression
            and building the kernel will be blocked.

        4,5,6) Same as #1,2,3, but with a phenos that has extra iids and for which the iids are shuffled.


        """

        # set up grid
        ##############################
        num_steps_delta = 5
        num_folds = 2

        # log_2 space and all SNPs
        k_values = [0, 1, 5, 10, 100, 500, 700, 10000]
        delta_values = np.logspace(
            -3, 3, endpoint=True, num=num_steps_delta, base=np.exp(1)
        )

        random_state = 42

        # case 1
        fss_1 = FeatureSelectionStrategy(
            snpreader,
            self.pheno_fn,
            num_folds,
            cov_fn=cov_fn,
            random_state=random_state,
            num_pcs=num_pcs,
            interpolate_delta=True,
            num_snps_in_memory=20000,
            count_A1=False,
        )
        best_k_1, best_delta_1, best_obj_1, best_snps_1 = fss_1.perform_selection(
            k_values,
            delta_values,
            output_prefix=output_prefix,
            select_by_ll=True,
            strategy=strategy,
            create_pdf=False,
        )

        # some misc testing
        from fastlmm.feature_selection import PerformSelectionDistributable as psd

        perform_selection_distributable = psd.PerformSelectionDistributable(
            fss_1,
            k_values,
            delta_values,
            strategy,
            output_prefix,
            select_by_ll=True,
            penalty=0.0,
            create_pdf=False,
        )
        self.assertEqual(perform_selection_distributable.work_count, 3)
        s = perform_selection_distributable.tempdirectory
        s = str(perform_selection_distributable)
        s = "%r" % perform_selection_distributable
        from fastlmm.feature_selection.feature_selection_cv import GClass

        s = "%r" % GClass.factory(snpreader, 1000000, Unit(), 50, count_A1=False)
        s = s
        #!!making  test for each break point.

        # case 2
        fss_2 = FeatureSelectionStrategy(
            snpreader,
            self.pheno_fn,
            num_folds,
            cov_fn=cov_fn,
            random_state=random_state,
            num_pcs=num_pcs,
            interpolate_delta=True,
            num_snps_in_memory=5000,
            count_A1=False,
        )
        best_k_2, best_delta_2, best_obj_2, best_snps_2 = fss_2.perform_selection(
            k_values,
            delta_values,
            output_prefix=output_prefix,
            select_by_ll=True,
            strategy=strategy,
            create_pdf=False,
        )

        # case 3
        fss_3 = FeatureSelectionStrategy(
            snpreader,
            self.pheno_fn,
            num_folds,
            cov_fn=cov_fn,
            random_state=random_state,
            num_pcs=num_pcs,
            interpolate_delta=True,
            num_snps_in_memory=600,
            count_A1=False,
        )
        best_k_3, best_delta_3, best_obj_3, best_snps_3 = fss_3.perform_selection(
            k_values,
            delta_values,
            output_prefix=output_prefix,
            select_by_ll=True,
            strategy=strategy,
            create_pdf=False,
        )

        # case 4
        fss_4 = FeatureSelectionStrategy(
            snpreader,
            self.pheno_shuffleplus_fn,
            num_folds,
            cov_fn=cov_fn,
            random_state=random_state,
            num_pcs=num_pcs,
            interpolate_delta=True,
            num_snps_in_memory=20000,
            count_A1=False,
        )
        best_k_4, best_delta_4, best_obj_4, best_snps_4 = fss_4.perform_selection(
            k_values,
            delta_values,
            output_prefix=output_prefix,
            select_by_ll=True,
            strategy=strategy,
            create_pdf=False,
        )

        # case 5
        fss_5 = FeatureSelectionStrategy(
            snpreader,
            self.pheno_shuffleplus_fn,
            num_folds,
            cov_fn=cov_fn,
            random_state=random_state,
            num_pcs=num_pcs,
            interpolate_delta=True,
            num_snps_in_memory=5000,
            count_A1=False,
        )
        best_k_5, best_delta_5, best_obj_5, best_snps_5 = fss_5.perform_selection(
            k_values,
            delta_values,
            output_prefix=output_prefix,
            select_by_ll=True,
            strategy=strategy,
            create_pdf=False,
        )

        # case 6
        fss_6 = FeatureSelectionStrategy(
            snpreader,
            self.pheno_shuffleplus_fn,
            num_folds,
            cov_fn=cov_fn,
            random_state=random_state,
            num_pcs=num_pcs,
            interpolate_delta=True,
            num_snps_in_memory=600,
            count_A1=False,
        )
        best_k_6, best_delta_6, best_obj_6, best_snps_6 = fss_6.perform_selection(
            k_values,
            delta_values,
            output_prefix=output_prefix,
            select_by_ll=True,
            strategy=strategy,
            create_pdf=False,
        )

        self.assertEqual(int(best_k_1), int(best_k_2))
        self.assertEqual(int(best_k_1), int(best_k_3))
        # self.assertEqual(int(best_k_1), int(best_k_4))
        # self.assertEqual(int(best_k_1), int(best_k_5))
        # self.assertEqual(int(best_k_1), int(best_k_6))
        self.assertAlmostEqual(best_obj_1, best_obj_2)
        self.assertAlmostEqual(best_obj_1, best_obj_3)
        # self.assertAlmostEqual(best_obj_1, best_obj_4)
        self.assertAlmostEqual(best_obj_4, best_obj_5)
        self.assertAlmostEqual(best_obj_4, best_obj_6)

        if strategy != "insample_cv":
            self.assertAlmostEqual(best_delta_1, best_delta_2)
            self.assertAlmostEqual(best_delta_1, best_delta_3)
            # self.assertAlmostEqual(best_delta_1, best_delta_4)
            self.assertAlmostEqual(best_delta_4, best_delta_5)
            self.assertAlmostEqual(best_delta_4, best_delta_6)

    def test_log_likelihood_bed(self):
        self.log_likelihood(self.snpreader_bed)

    def test_log_likelihood_hdf5(self):
        self.log_likelihood(self.snpreader_hdf5)

    def test_log_likelihood_dat(self):
        self.log_likelihood(self.snpreader_dat)

    def test_log_likelihood_ped(self):
        self.log_likelihood(self.snpreader_ped)

    def log_likelihood(self, snpreader):
        """
        test mean, variance against C++ results (autoselect):

        FastLmmC.exe -autoselect test -bfilesim toydata -pheno toydata.phe -logDelta X -autoSelectSearchValues k -verboseOutput > debug.txt
        """

        # low-rank, small delta
        ll_1_expected = np.array(
            [
                79.7217,
                80.5289,
                79.9218,
                53.7485,
                71.4134,
                71.5751,
                58.4209,
                69.82,
                85.5727,
                72.7218,
            ]
        )
        ll_1 = core_run(snpreader, self.pheno_fn, 50, np.exp(-5))

        for i in range(len(ll_1)):
            np.testing.assert_approx_equal(
                ll_1[i],
                ll_1_expected[i],
                significant=3,
                err_msg="Log-likelihoods differ",
                verbose=True,
            )

        # low-rank, large delta
        ll_2_expected = np.array(
            [
                68.7098,
                70.0446,
                75.1816,
                62.3675,
                69.34,
                74.6755,
                59.9937,
                66.6408,
                74.1564,
                68.3146,
            ]
        )
        ll_2 = core_run(snpreader, self.pheno_fn, 50, np.exp(10))

        for i in range(len(ll_2)):
            np.testing.assert_approx_equal(
                ll_2[i],
                ll_2_expected[i],
                significant=3,
                err_msg="Log-likelihoods differ",
                verbose=True,
            )

        # full-rank, small delta
        # these values appear to indicate numerical instability. the python version will continue to match these
        # results for the time being, but provide a flag "robust" to lead to a numerically more stable solution
        ll_3_expected = np.array(
            [
                1636.33,
                28711.8,
                32008.8,
                1363.74,
                128444,
                22277.6,
                16389.2,
                95458.7,
                4710.33,
                68308.9,
            ]
        )
        ll_3 = core_run(snpreader, self.pheno_fn, 5000, np.exp(-5))

        for i in range(len(ll_3)):
            np.testing.assert_approx_equal(
                ll_3[i],
                ll_3_expected[i],
                significant=2,
                err_msg="Log-likelihoods differ",
                verbose=True,
            )

        # full-rank, small delta
        ll_4_expected = np.array(
            [
                68.4794,
                70.7483,
                76.6886,
                62.2721,
                69.0659,
                76.8207,
                59.5216,
                66.1517,
                75.9149,
                68.6061,
            ]
        )
        ll_4 = core_run(snpreader, self.pheno_fn, 5000, np.exp(10))

        for i in range(len(ll_4)):
            np.testing.assert_approx_equal(
                ll_4[i],
                ll_4_expected[i],
                significant=3,
                err_msg="Log-likelihoods differ",
                verbose=True,
            )

        # square matrix, delta 2
        ll_5_expected = np.array(
            [
                259.91,
                351.914,
                389.971,
                313.807,
                334.883,
                345.312,
                221.556,
                206.15,
                270.53,
                304.231,
            ]
        )
        ll_5 = core_run(snpreader, self.pheno_fn, 500, np.exp(2))

        for i in range(len(ll_5)):
            np.testing.assert_approx_equal(
                ll_5[i],
                ll_5_expected[i],
                significant=3,
                err_msg="Log-likelihoods differ",
                verbose=True,
            )


def core_run(snpreader, pheno_fn, k, delta):
    """
    extracted core functionality, to avoid shuffle of data and not correct delta
    """
    with patch.dict("os.environ", {"ARRAY_MODULE": "numpy"}) as _:

        G, X, y = load_snp_data(snpreader, pheno_fn, standardizer=Unit())
        kf = KFold(n_splits=10, shuffle=False).split(list(range(len(y))))

        ll = np.zeros(10)

        fold_idx = 0
        fold_data = {}
        for split_idx, (train_idx, test_idx) in enumerate(kf):
            fold_idx += 1

            fold_data["train_idx"] = train_idx
            fold_data["test_idx"] = test_idx

            # set up data
            ##############################
            fold_data["G_train"] = G[train_idx, :].read()
            fold_data["G_test"] = G[test_idx, :]

            fold_data["X_train"] = X[train_idx]
            fold_data["X_test"] = X[test_idx]

            fold_data["y_train"] = y[train_idx]
            fold_data["y_test"] = y[test_idx]

            # feature selection
            ##############################
            _F, _pval = lin_reg.f_regression_block(
                lin_reg.f_regression_cov_alt,
                fold_data["G_train"].val,
                fold_data["y_train"],
                blocksize=1e4,
                C=fold_data["X_train"],
            )
            feat_idx = np.argsort(_pval)
            fold_data["feat_idx"] = feat_idx

            # re-order SNPs (and cut to max num)
            ##############################
            fold_data["G_train"] = fold_data["G_train"][:, feat_idx[0:k]].read()
            fold_data["G_test"] = fold_data["G_test"][:, feat_idx[0:k]].read()

            model = getLMM()
            model.setG(fold_data["G_train"].val)
            model.sety(fold_data["y_train"])
            model.setX(fold_data["X_train"])

            REML = False

            # predict on test set
            res = model.nLLeval(delta=delta, REML=REML)
            model.setTestData(Xstar=fold_data["X_test"], G0star=fold_data["G_test"].val)
            model.predictMean(beta=res["beta"], delta=delta)
            # mse_cv1[k_idx, delta_idx] = mean_squared_error(fold_data["y_test"],
            # out)
            ll[split_idx] = model.nLLeval_test(
                fold_data["y_test"], res["beta"], sigma2=res["sigma2"], delta=delta
            )

        return ll


def getTestSuite():
    """
    set up composite test suite
    """
    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestTwoKernelFeatureSelection)
    suite2 = unittest.TestLoader().loadTestsFromTestCase(TestFeatureSelection)
    return unittest.TestSuite([suite1, suite2])


if __name__ == "__main__":
    unittest.main()



import numpy as np
import scipy.linalg as la


def gendat(
    TWO_KERNEL,
    nInd,
    nSnp,
    nCovar,
    minMaf=0.05,
    maxMaf=0.4,
    minSigE2=0.5,
    maxSigE2=1,
    minSigG2=0.5,
    maxSigG2=1,
):
    """
    Generate synthetic SNPs and phenotype.
    SNPs are iid, and there is no population structure.
    Phenotype y is generated from a LMM with SNPs in a PS kernel.

    Returns:
        covDat
        y
        psSnps
    """

    if TWO_KERNEL:
        psSnps = gensnps(nInd, nSnp, minMaf, maxMaf)
        psK = psSnps.dot(psSnps.T)
        psK += 1e-5 * np.eye(nInd)
        psKchol = la.cholesky(psK)
    else:
        psSnps = None

    covDat = np.random.uniform(0, 1, (nInd, nCovar))
    covWeights = np.random.uniform(-0.5, 0.5, (nCovar, 1))

    sigE2 = np.random.uniform(low=minSigE2, high=maxSigE2)
    sigG2 = np.random.uniform(low=minSigG2, high=maxSigG2)

    ##generate the phenotype using the background kernel and covariates
    if TWO_KERNEL:
        y_pop = np.sqrt(sigG2) * psKchol.dot(np.randn(nInd, 1))
    else:
        y_pop = 0
    y_noise = np.randn(nInd, 1) * np.sqrt(sigE2)
    y = (covDat.dot(covWeights) + y_pop + y_noise).squeeze()
    return covDat, y, psSnps


def gensnps(numSnp, nInd, minMaf=0.05, maxMaf=0.4, standardize=False):
    """
    Generate independent SNPs, from independent people (no population structure)
    """

    if (minMaf < 0) | (minMaf > 1):
        raise Exception("invalid minMaf provided")
    if (maxMaf < 0) | (maxMaf > 1):
        raise Exception("invalid minMaf provided")
    if maxMaf < minMaf:
        raise Exception("invalid minMaf/maxMaf combo provided")

    maf = np.random.uniform(low=minMaf, high=maxMaf, size=(numSnp, 1))

    numChrm = 2
    snpDat = np.zeros((numSnp, nInd))
    for ch in np.arange(0, numChrm):
        mafRep = np.tile(maf, (1, nInd))
        snpDat += np.random.uniform(size=(numSnp, nInd)) < mafRep

    if standardize:
        raise NotImplementedError()

    return snpDat



import platform
import os
import shutil
from setuptools import setup
import numpy

# Work around https://github.com/pypa/pip/issues/7953
import site
import sys

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

# Version number
version = "0.6.8"


def readme():
    with open("README.md") as f:
        return f.read()


# python setup.py sdist bdist_wininst upload
setup(
    name="fastlmm",
    version=version,
    description="Fast GWAS",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="gwas bioinformatics LMMs MLMs linear mixed models genomics genetics python",
    url="https://fastlmm.github.io/",
    author="FaST-LMM Team",
    author_email="fastlmm-dev@python.org",
    project_urls={
        "Bug Tracker": "https://github.com/fastlmm/FaST-LMM/issues",
        "Documentation": "http://fastlmm.github.io/FaST-LMM",
        "Source Code": "https://github.com/fastlmm/FaST-LMM",
    },
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
    ],
    packages=[  # basically, everything with a __init__.py
        "fastlmm",
        "fastlmm/association",
        "fastlmm/association/altset_list",
        "fastlmm/association/tests",
        "fastlmm/external",
        "fastlmm/external/util",
        "fastlmm/feature_selection",
        "fastlmm/inference",
        "fastlmm/inference/tests",
        "fastlmm/pyplink",  # old snpreader
        "fastlmm/pyplink/altset_list",  # old snpreader
        "fastlmm/pyplink/snpreader",  # old snpreader
        "fastlmm/pyplink/snpset",  # old snpreader
        "fastlmm/util",
        "fastlmm/util/matrix",
        "fastlmm/util/standardizer",
        "fastlmm/util/stats",
    ],
    package_data={
        "fastlmm/association": [
            "Fastlmm_autoselect/FastLmmC.exe",
            "Fastlmm_autoselect/libiomp5md.dll",
            "Fastlmm_autoselect/fastlmmc",
            "Fastlmm_autoselect/FastLmmC.Manual.pdf",
        ],
        "fastlmm/feature_selection": [
            "examples/bronze.txt",
            "examples/ScanISP.Toydata.config.py",
            "examples/ScanLMM.Toydata.config.py",
            "examples/ScanOSP.Toydata.config.py",
            "examples/toydata.5chrom.bed",
            "examples/toydata.5chrom.bim",
            "examples/toydata.5chrom.fam",
            "examples/toydata.cov",
            "examples/toydata.map",
            "examples/toydata.phe",
            "examples/toydata.shufflePlus.phe",
            "examples/toydata.sim",
            "examples/toydataTest.phe",
            "examples/toydataTrain.phe",
        ],
        "fastlmm": ["util/fastlmm.hashdown.json"],
    },
    install_requires=[
        "pandas>=1.1.1",
        "matplotlib>=1.5.1",
        "scikit-learn>=0.19.1",
        "bed-reader>=1.0.4",
        "pysnptools>=0.5.11",
        "cloudpickle>=2.2.0",
        "statsmodels>=0.10.1",
        "psutil>=5.6.7",
        "fastlmmclib>=0.0.3",
    ],
)
